from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid
import bpy


class ULGetMaterialNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.mat_name = None
        self.node_name = None
        self.OUT = ULOutSocket(self, self._get_val)

    def _get_val(self):
        mat_name = self.get_input(self.mat_name)
        node_name = self.get_input(self.node_name)
        if is_invalid(mat_name, node_name):
            return STATUS_WAITING
        return (
            bpy.data.materials[mat_name]
            .node_tree
            .nodes[node_name]
        )

    def evaluate(self):
        self._set_ready()



from bge import constraints
from bge import logic
from bge import render
from bge.types import KX_ConstraintWrapper as GameConstraint
from bge.types import KX_GameObject as GameObject
from mathutils import Vector
from uplogic.utils import debug, get_direction
from uplogic.utils import set_curve_points
from uplogic.utils import xrot_to
from uplogic.utils import yrot_to
from uplogic.utils import zrot_to


CONSTRAINT_TYPES = {
    'point': 0,
    'hinge': 1,
    'angular': 2,
    'conetwist': 3,
    'generic6dof': 4
}


def create_constraint(
    obj: GameObject,
    target: GameObject,
    constraint_type: int or str = 0,
    pivot: set = (0, 0, 0),
    limit: set = (0, 0, 0),
    linked_collision: bool = True,
    local: bool = True
) -> GameConstraint:
    """Wrapper function for `bge.constraints.createConstraint()`. Creates a constraint

    :param `obj`: Object the constraint will be applied to.
    :param `target`: Target for the constraint.
    :param `constraint_type`: Type of the constraint. One of `['point', 'hinge', 'angular', 'conetwist', 'generic6dof']`.
    :param `pivot`: Point of application for the constraint.
    :param `limit`: Limit movement of the object (Like a doorstop).
    :param `linked_collision`: Enable/Disable collision between obj and target.
    :param `local`: Use obj's local space.
    """
    if not local:
        pivot[0] -= obj.worldPosition.x
        pivot[1] -= obj.worldPosition.y
        pivot[2] -= obj.worldPosition.z
    return constraints.createConstraint(
        obj.getPhysicsId(),
        target.getPhysicsId(),
        constraint_type if isinstance(constraint_type, int) else CONSTRAINT_TYPES.get(constraint_type, 0),
        pivot[0],
        pivot[1],
        pivot[2],
        limit[0],
        limit[1],
        limit[2],
        0 if linked_collision else 128
    )


def remove_constraint(constraint: GameConstraint) -> None:
    """Wrapper function for `bge.constraints.removeConstraint()`. Creates a constraint

    :param `constraint`: The constraint to remove.
    """
    constraints.removeConstraint(constraint.getConstraintId())


class ULTrackTo():

    _deprecated = True

    def __init__(
        self,
        game_object: GameObject,
        target: GameObject or Vector,
        axis: int = 2,
        front: int = 1,
        factor: float = 1
    ) -> None:
        if self._deprecated:
            debug('ULTrackTo class will be renamed to "TrackTo" in future releases!')
        self._axis = None
        self._target = None
        self.game_object = game_object
        self.target = target
        self.front = front
        self.factor = factor
        self.axis = axis

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, val):
        if isinstance(val, list) or isinstance(val, tuple):
            self._target = Vector(val)
        elif isinstance(val, Vector):
            self._target = val
        else:
            debug('Could not set TrackTo target!')

    @property
    def axis(self):
        return self._axis

    @axis.setter
    def axis(self, val):
        if val == 0:
            self.rotate_func = xrot_to
        elif val == 1:
            self.rotate_func = yrot_to
        elif val == 2:
            self.rotate_func = zrot_to
        else:
            self.rotate_func = None
        self._axis = val
        logic.getCurrentScene().pre_draw.append(self.update)

    def remove(self):
        logic.getCurrentScene().pre_draw.remove(self.update)

    def update(self):
        if self.rotate_func:
            self.rotate_func(self.game_object, self.target, self.front, self.speed)


class TrackTo(ULTrackTo):
    _deprecated = False


class ULSpring():
    """Spring Physics Constraint. The two objects connected by the string will
    be pulled towards each other if the string is streched, optionally they will
    be pushed apart when the spring is being compacted.

    :param `origin`: First connection point of the spring.
    :param `target`: Second connection point of the spring.
    :param `rigid_body_origin`: Object to be influenced by the spring (optional).
    :param `rigid_body_target`: Object to be influenced by the spring (optional).
    :param `stiffness`: Amount the spring will bounce back.
    :param `max_force`: Maximum force the spring will use.
    :param `use_push`: Push the objects apart when spring is compressed.
    :param `use_breaking`: Remove the constraint when spring is pulled too much.
    :param `break_threshold`: Amount of strain the spring will endure.
    :param `curve`: Set a curve object to fit the spring.
    :param `visualize`: Enable a visual representation of the spring.
    """

    _deprecated = True

    def __init__(
        self,
        origin: GameObject,
        target: GameObject,
        rigid_body_origin: GameObject = None,
        rigid_body_target: GameObject = None,
        stiffness: float = 1,
        max_force: float = -1,
        distance: float = None,
        use_push: bool = False,
        use_breaking: bool = False,
        break_threshold: float = 1,
        curve: GameObject or None = None,
        visualize: bool = False
    ) -> None:
        if self._deprecated:
            debug('ULSpring class will be renamed to "Spring" in future releases!')
        self.force = 0
        if isinstance(origin, tuple) or isinstance(origin, list):
            origin = Vector((origin))
        self.origin = origin
        if isinstance(target, tuple) or isinstance(target, list):
            target = Vector((target))
        self.target = target
        self.use_push = use_push
        self.rigid_body_origin = rigid_body_origin if rigid_body_origin else origin
        self.rigid_body_target = rigid_body_target if rigid_body_target else target
        self.stiffness = stiffness
        self.max_force = max_force
        self.use_breaking = use_breaking
        self.break_threshold = break_threshold
        self.visualize = visualize
        obj_dist = origin.getDistanceTo(target)
        self.distance = distance if distance is not None else obj_dist
        self.curve = curve
        if not use_breaking or self.distance >= obj_dist:
            self.update()
            logic.getCurrentScene().pre_draw.append(self.update)

    @property
    def points(self):
        return [self.origin.worldPosition, self.target.worldPosition]
    
    @points.setter
    def points(self, val):
        print("Attribute 'points' is read-only")

    @property
    def active(self):
        return self.force != 0

    @active.setter
    def active(self, val):
        print("Attribute 'active' is read-only")

    def remove(self):
        pre_draw = logic.getCurrentScene().pre_draw
        if self.update in pre_draw:
            pre_draw.remove(self.update)

    def update(self):
        o = self.origin
        t = self.target
        force = (o.getDistanceTo(t) - self.distance) * self.stiffness
        if self.max_force >= 0 and force > self.max_force:
            force = self.max_force
        if not self.use_push:
            force = force if force >= 0 else 0
        if self.use_breaking and force > self.break_threshold:
            logic.getCurrentScene().pre_draw.remove(self.update)
            return
        if self.visualize:
            start = getattr(o, 'worldPosition', o)
            end = getattr(t, 'worldPosition', t)
            render.drawLine(
                start,
                end,
                # [1, 1-abs(power), 1-abs(power)]
                [abs(force), 0, 1-abs(force)]
            )
        self.force = force
        if self.curve:
            set_curve_points(self.curve, self.points)
        rbo = self.rigid_body_origin
        rbt = self.rigid_body_target
        if hasattr(rbo, 'blenderObject') and rbo.blenderObject.data:
            rbo.applyImpulse(o.worldPosition, get_direction(o, t) * force)
        if hasattr(rbt, 'blenderObject') and rbt.blenderObject.data:
            rbt.applyImpulse(o.worldPosition, get_direction(t, o) * force)


class Spring(ULSpring):
    _deprecated = False



from bge.types import KX_GameObject as GameObject
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting


class ULChildByIndex(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.from_parent: GameObject = None
        self.index: int = None
        self.CHILD = ULOutSocket(self, self.get_child)

    def get_child(self):
        socket = self.get_output('child')
        if socket is None:
            parent: GameObject = self.get_input(self.from_parent)
            index: int = self.get_input(self.index)
            if is_waiting(parent, index):
                return STATUS_WAITING
            elif not is_invalid(parent):
                if len(parent.children) > index:
                    return self.set_output(
                        'child',
                        parent.children[index]
                    )
            return STATUS_WAITING
        return socket

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.input import VRHeadset


class ULGetVRHeadsetValues(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.controller = VRHeadset()
        self.POS = ULOutSocket(self, self.get_pos)
        self.ORI = ULOutSocket(self, self.get_ori)

    def get_pos(self):
        return self.controller.position
    
    def get_ori(self):
        return self.controller.orientation.to_euler()

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting


class ULExtendList(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.list_1: list = None
        self.list_2: list = None
        self.LIST = ULOutSocket(self, self.get_list)

    def get_list(self):
        socket = self.get_output('list')
        if socket is None:
            list_1 = self.get_input(self.list_1)
            list_2 = self.get_input(self.list_2)
            if is_waiting(list_1, list_2):
                return STATUS_WAITING
            return self.set_output('list', list_1.extend(list_2))
        return socket

    def evaluate(self):
        self._set_ready()


from bge import render
from uplogic.input import MOUSE
from mathutils import Vector


class HoverBehavior():
    @property
    def hover(self):
        screen_size = [render.getWindowWidth(), render.getWindowHeight()]
        mpos = Vector((MOUSE.position.x, 1 - MOUSE.position.y))
        dsize = self._draw_size
        return (
            self.pos_abs[0] <= mpos.x * screen_size[0] <= self.pos_abs[0] + dsize[0] and
            self.pos_abs[1] <= mpos.y * screen_size[1] <= self.pos_abs[1] + dsize[1] and
            not self.canvas._hover_consumed
        )



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.ui import Canvas
from uplogic.utils import not_met


class ULAddUIWidget(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.parent_widget = None
        self.child_widget = None
        self._done = False
        self.OUT = ULOutSocket(self, self._get_done)

    def _get_done(self):
        return self._done

    def evaluate(self):
        self._done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        self._set_ready()
        self.get_input(self.parent_widget).add_widget(self.get_input(self.child_widget))
        self._done = True
        



from bge import logic
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import not_met
from uplogic.utils import is_waiting
from uplogic.utils import debug


class ULGamepadVibration(ULActionNode):
    def __init__(self, axis=0):
        ULActionNode.__init__(self)
        self.condition = None
        self.index = None
        self.left = None
        self.right = None
        self.time = None
        self.done = None
        self.DONE = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self._set_ready()
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        index = self.get_input(self.index)
        left = self.get_input(self.left)
        right = self.get_input(self.right)
        time = self.get_input(self.time)
        if is_waiting(index, left, right, time):
            return

        if not logic.joysticks[index]:
            return
        joystick = logic.joysticks[index]
        if not joystick.hasVibration:
            debug('Joystick at index {} has no vibration!'.format(index))
            return

        joystick.strengthLeft = left
        joystick.strengthRight = right
        joystick.duration = int(round(time * 1000))

        joystick.startVibration()
        self.done = True



from .widget import Widget
from .behaviors import HoverBehavior
from .label import Label
import gpu
from uplogic.input.mouse import MOUSE_EVENTS, LMB, RMB
from uplogic.utils import debug


class Button(Widget, HoverBehavior):

    def __init__(self, pos=[0., 0.], size=[100., 100.], bg_color=(0, 0, 0, 0), relative={}, border_width=1.0, border_color=(0, 0, 0, 0), hover_color=(0, 0, 0, .5), halign='left', valign='bottom'):
        super().__init__(pos, size, bg_color, relative, halign=halign, valign=valign)
        self.hover_color = hover_color
        self.border_width = border_width
        self.border_color = border_color
        self._clicked = False
        self._released = False
        self._in_focus = False
        self._down = False

    def draw(self):
        super()._setup_draw()
        if self.hover:
            self._in_focus = True
            self.on_hover(self)
            self.canvas._hover_consumed = True
        else:
            self._in_focus = False

        self._released = False
        self._clicked = False
        gpu.state.line_width_set(self.border_width)
        gpu.state.point_size_set(self.border_width)
        self._shader.uniform_float("color", self.hover_color if self._in_focus else self.bg_color)
        self._batch.draw(self._shader)
        self._shader.uniform_float("color", self.border_color)
        self._batch_line.draw(self._shader)
        self._batch_points.draw(self._shader)
        super().draw()
        if self._in_focus and MOUSE_EVENTS[LMB].active and not self.canvas._click_consumed:
            self.on_click(self)
            self._clicked = True
            self.canvas._click_consumed = True
            self._down = True
        elif not MOUSE_EVENTS[LMB].active and self._down:
            self.on_release(self)
            self._down = False
            self._released = True
            self.canvas._click_consumed = False
        elif self._down:
            self.on_hold(self)

    def on_click(self, widget):
        pass
        # debug(f'{self} clicked.')

    def on_hold(self, widget):
        pass
        # debug(f'{self} is being held.')

    def on_release(self, widget):
        pass
        # debug(f'{self} is released.')

    def on_hover(self, widget):
        pass
        # debug(f'{self} is hovered over.')


class LabelButton(Button, HoverBehavior):

    def __init__(
        self,
        pos=[0., 0.],
        size=[100., 100.],
        relative={},
        bg_color=(0, 0, 0, 0),
        border_width=1.0,
        border_color=(0, 0, 0, 0),
        hover_color=(0, 0, 0, .5),
        text='',
        text_pos=[.5, .5],
        font='',
        font_size=12,
        font_color=(1, 1, 1, 1),
        line_height=1.5,
        halign='left',
        valign='bottom',
        halign_text='center',
        valign_text='center'
    ):
        super().__init__(
            pos,
            size,
            bg_color,
            relative,
            border_width=border_width,
            border_color=border_color,
            hover_color=hover_color,
            halign=halign,
            valign=valign
        )
        self.label = Label(
            relative={'pos': True},
            pos=text_pos,
            font=font,
            halign=halign_text,
            valign=valign_text,
            text=text,
            font_color=font_color,
            font_size=font_size,
            line_height=line_height
        )
        self.add_widget(self.label)
        self._in_focus = False

    @property
    def text(self):
        return self.label.text

    @text.setter
    def text(self, val):
        self.label.text = val

    @property
    def text_pos(self):
        return self.label.pos

    @text_pos.setter
    def text_pos(self, val):
        self.label.pos = val

    @property
    def font(self):
        return self.label.font

    @font.setter
    def font(self, val):
        self.label.font = val

    @property
    def font_size(self):
        return self.label.font_size

    @text.setter
    def font_size(self, val):
        self.label.font_size = val

    @property
    def font_color(self):
        return self.label.font_color

    @font_color.setter
    def font_color(self, val):
        self.label.font_color = val

    @property
    def line_height(self):
        return self.label.line_height

    @line_height.setter
    def line_height(self, val):
        self.label.line_height = val

    @property
    def halign_text(self):
        return self.label.halign

    @halign_text.setter
    def halign_text(self, val):
        self.label.halign = val

    @property
    def valign_text(self):
        return self.label.valign

    @valign_text.setter
    def valign_text(self, val):
        self.label.valign = val



from bge import logic
from bge.types import KX_GameObject as GameObject
from math import pi
from mathutils import Vector
from uplogic.utils import debug
from uplogic.utils import interpolate
from uplogic.events import schedule_callback


XBOX = {
    'A': 0,
    'B': 1,
    'X': 2,
    'Y': 3,
    'SELECT': 4,
    'BACK': 4,
    'START': 6,
    'MENU': 6,
    'LS': 7,
    'L3': 7,
    'RS': 8,
    'R3': 8,
    'LB': 9,
    'RB': 10,
    'DPADUP': 11,
    'DPADDOWN': 12,
    'DPADLEFT': 13,
    'DPADRIGHT': 14,
    'RT': 5,
    'LT': 4
}
"""Buttons in [`A`, `B`, `X`, `Y`, `SELECT`, `BACK`, `START`, `MENU`, `LS`,
`L3`, `RS`, `R3`, `LB`, `RB`, `DPADUP`, `DPADDOWN`, `DPADLEFT`, `DPADRIGHT`,
`RT`, `LT`]"""


SONY = {
    'X': 0,
    'CROSS': 0,
    'CIRCLE': 1,
    'SQUARE': 2,
    'TRIANGLE': 3,
    'SELECT': 4,
    'SHARE': 4,
    'START': 6,
    'MENU': 6,
    'L3': 7,
    'R3': 8,
    'L1': 9,
    'R1': 10,
    'DPADUP': 11,
    'DPADDOWN': 12,
    'DPADLEFT': 13,
    'DPADRIGHT': 14,
    'R2': 5,
    'L2': 4
}
"""Buttons in [`X`, `CROSS`, `CIRCLE`, `SQUARE`, `TRIANGLE`, `SELECT`, `SHARE`,
`START`, `MENU`, `L3`, `R3`, `L1`, `R1`, `DPADUP`, `DPADDOWN`, `DPADLEFT`,
`DPADRIGHT`, `R2`, `L2`]"""


LS = 'LS'
RS = 'RS'

STICKS = {
    LS: [0, 1],
    RS: [2, 3]
}


_active_buttons = {}
_active_axis = {}


def gamepad_button(
    button: int,
    idx: int = 0,
    tap: int = False,
    released: int = False
) -> bool:
    '''Retrieve button value.\n
    Not intended for manual use.
    '''
    global _active_buttons
    if logic.joysticks[idx] is None:
        return False
    gamepad = logic.joysticks[idx]
    state = button in gamepad.activeButtons
    if tap or released:
        if released:
            tap_cond = _active_buttons.get(button, False)
            _active_buttons[button] = state
            return not state and tap_cond
        tap_cond = not _active_buttons.get(button, False)
        _active_buttons[button] = state
        return state and tap_cond
    _active_buttons[button] = state
    return state


def gamepad_axis(
    axis: int,
    idx: int = 0,
    tap: bool = False,
    released: bool = False,
    threshold: float = .07
) -> float:
    '''Retrieve axis value.\n
    Not intended for manual use.
    '''
    global _active_axis
    if logic.joysticks[idx] is None:
        return 0.0
    gamepad = logic.joysticks[idx]
    if released:
        val = gamepad.axisValues[axis]
        if _active_axis.get(axis, 0) != 0 and val == 0:
            _active_axis[axis] = val
            return 1.0
        else:
            _active_axis[axis] = val
            return 0.0
    if tap:
        if _active_axis.get(axis, 0) == 0:
            val = gamepad.axisValues[axis]
            _active_axis[axis] = val
            return val if abs(val) >= threshold else 0
        else:
            _active_axis[axis] = gamepad.axisValues[axis]
            return 0.0
    val = gamepad.axisValues[axis]
    _active_axis[axis] = val
    return val if abs(val) >= threshold else 0


def gamepad_trigger(
    trigger: str = 'LT',
    idx: int = 0,
    threshold: float = .07
) -> float:
    """Retrieve gamepad trigger values.

    :param `trigger`: Whether to use the left or right trigger;
    `str` in [`'LT'`, `'RT'`].
    :param `idx`: index of the gamepad.
    :param `threshold`: Only detect values higher than the threshold.

    :returns: Intensity of the selected trigger.
    """
    return gamepad_axis(4 if trigger == 'LT' else 5, idx, threshold=threshold)


def gamepad_stick(
    stick: str = LS,
    idx: int = 0,
    threshold: float = .07
) -> set:
    '''Retrieve stick values.

    :param stick: which stick to use.
    can bei either `LS` or `RS` from `uplogic.input`.
    :param idx: gamepad index (default = 0).
    :param threshold: minimum value for each axis to be valid.

    :returns: set `(x, y)`
    '''
    xaxis = STICKS[stick][0]
    yaxis = STICKS[stick][1]
    return Vector((
        gamepad_axis(xaxis, idx, threshold=threshold),
        gamepad_axis(yaxis, idx, threshold=threshold)
    ))


def gamepad_tap(
    button: str,
    idx: int = 0,
    layout: dict = XBOX
) -> float or bool:
    '''Detect button tap.

    :param button: button name as `str` (e.g. `'START'`)
    :param idx: gamepad index (default = 0).
    :param layout: gamepad layout,
    can be either `XBOX` or `SONY` from `uplogic.input`.

    :returns: float or boolean
    '''
    btn_idx = layout[button]
    if button in ['R2', 'L2', 'RT', 'LT']:
        return gamepad_axis(btn_idx, idx, True)
    else:
        return gamepad_button(btn_idx, idx, True)


def gamepad_down(
    button: str,
    idx: int = 0,
    layout: dict = XBOX
) -> float or bool:
    '''Detect button held down.

    :param button: button name as `str` (e.g. `'START'`)
    :param idx: gamepad index (default = 0).
    :param layout: gamepad layout,
    can be either `XBOX` or `SONY` from `uplogic.input`.

    :returns: float or boolean
    '''
    btn_idx = layout[button]
    if button in ['R2', 'L2', 'RT', 'LT']:
        return gamepad_axis(btn_idx, idx)
    else:
        return gamepad_button(btn_idx, idx)


def gamepad_up(
    button: str,
    idx: int = 0,
    layout: dict = XBOX
) -> float or bool:
    '''Detect button released.

    :param button: button name as `str` (e.g. `'START'`)
    :param idx: gamepad index (default = 0).
    :param layout: gamepad layout,
    can be either `XBOX` or `SONY` from `uplogic.input`.

    :returns: float or boolean
    '''
    btn_idx = layout[button]
    if button in ['R2', 'L2', 'RT', 'LT']:
        return gamepad_axis(btn_idx, idx, True, True)
    else:
        return gamepad_button(btn_idx, idx, True, True)


def gamepad_vibrate(idx: int = 0, strength: tuple = (.5, .5), time: float = 1.0):
    """Start the vibrators of the gamepad if available.

    :param `idx`: 
    """
    joystick = logic.joysticks[idx]
    if not joystick or not joystick.hasVibration:
        debug(f'Joystick at index {idx} has no vibration!')
    joystick.strengthLeft = strength[0]
    joystick.strengthRight = strength[1]
    joystick.duration = int(round(time * 1000))

    joystick.startVibration()


class ULGamePad():
    """Wrapper class for a controller. The index determines which connected
    gamepad to use, the layout determines whether to use XBox or Sony button
    naming.
    
    :param `idx`: Which gamepad to use.
    :param `layout`: Layout determining button names; `str` in [`'XBOX'`,
    `'SONY'`].
    """

    _deprecated = True

    def __init__(
        self,
        idx: int = 0,
        layout: dict = XBOX
    ) -> None:
        if self._deprecated:
            debug('ULGamePad class will be renamed to "Gamepad" in future releases!')
        self.idx = idx
        self.layout = layout
        if not logic.joysticks[idx]:
            debug(f'No Joystick found at index: {idx}')
        self.joystick = logic.joysticks[idx]

    def button_down(self, button: str):
        return gamepad_down(button, self.idx, self.layout)
        
    def button_tap(self, button: str):
        return gamepad_tap(button, self.idx, self.layout)

    def button_up(self, button: str):
        return gamepad_up(button, self.idx, self.layout)

    def sticks(self, stick: str = LS, threshold: float = 0.07):
        return gamepad_stick(stick, self.idx, threshold)
    
    def vibrate(self, strength: tuple = (.5, .5), time: float = 1.0):
        if not self.joystick.hasVibration:
            debug('Joystick at index {} has no vibration!'.format(self.idx))
            return
        self.joystick.strengthLeft = strength[0]
        self.joystick.strengthRight = strength[1]
        self.joystick.duration = int(round(time * 1000))

        self.joystick.startVibration()


class Gamepad(ULGamePad):
    _deprecated = False


class ULGamepadLook():
    """Automatically track the mouse movement and translate it to a rotate a
    body and optionally a head.

    This component can be activated/deactivated at any time to keep performance
    up.
    
    :param `obj`: Main object to rotate around the object's Z axis.
    :param `head`: Head object to rotate around the object's X/Y axis.
    :param `sensitivity`: Translation factor of mouse movement to rotation.
    :param `use_cap_x`: Whether to use capping on the mouse X movement (Z axis
    rotation).
    :param `cap_x`: Minimum and Maximum amount of rotation on the Z axis.
    :param `use_cap_y`: Whether to use capping on the mouse Y movement (X/Y axis
    rotation).
    :param `cap_y`: Minimum and Maximum amount of rotation on the X/Y axis.
    :param `invert`: Whether to use inverted values for mous X/Y movement.
    :param `smoothing`: Amount of movement smoothing.
    :param `local`: Whether to use local transform for the body object.
    :param `front`: Front axis (traditionally in blender, Y is front).
    :param `active`: Whether to start this component in active or inactive mode
    (can be changed later).
    """
    _deprecated = True
    def __init__(
        self,
        obj: GameObject,
        head: GameObject = None,
        sensitivity: float = .05,
        use_cap_x: bool = False,
        cap_x: tuple = (0, 0),
        use_cap_y: bool = False,
        cap_y: tuple = (-89, 89),
        invert: tuple = (True, True),
        smoothing: float = 0.0,
        local: bool = True,
        front: int = 1,
        idx: int = 0,
        stick: int = 'RS',
        threshold: float = 0.07,
        exponent: float = 2.3,
        active=True
    ) -> None:
        if self._deprecated:
            debug('ULGamepadLook class will be renamed to "GamepadLook" in future releases!')
        self.obj = obj
        self.head = head if head else obj
        self._defaults = [
            obj.localOrientation.copy(),
            head.localOrientation.copy()
            if head else
            obj.localOrientation.copy()]
        self.sensitivity = sensitivity
        self.use_cap_x = use_cap_x
        self.cap_x = cap_x
        self.use_cap_y = use_cap_y
        self.cap_y = cap_y
        self.invert = invert
        self.smoothing = smoothing
        self.initialized = False
        self.front = front
        self._x = 0
        self._y = 0
        self._active = False
        self.local = local
        self.reset_factor = 0
        self.stick = stick
        self.threshold = threshold
        self.exponent = exponent
        self.joystick = logic.joysticks[idx]
        self.active = active

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, val):
        scene = logic.getCurrentScene()
        if val and self.update not in scene.pre_draw and self.joystick:
            scene.pre_draw.append(self.update)
        elif not val and self.update in scene.pre_draw:
            self.initialized = False
            scene.pre_draw.remove(self.update)
        self._active = val

    @property
    def rotation(self):
        return self.obj.worldOrientation, self.head.worldOrientation

    @rotation.setter
    def rotation(self, val):
        self.obj.worldOrientation = val[0]
        self.head.worldOrientation = val[1]

    def stop(self):
        self.active = False
        self.initialized = False

    def disable(self):
        self.active = False

    def enable(self):
        self.active = True

    def reset(self, factor=1):
        if factor < 1:
            self.active = False
            if self.reset_factor < 1:
                self.obj.localOrientation = self.obj.localOrientation.lerp(self._defaults[0], factor)
                self.head.localOrientation = self.head.localOrientation.lerp(self._defaults[1], factor)
                self.reset_factor = interpolate(self.reset_factor, 1, factor)
                schedule_callback(self.reset, arg=factor)
            else:
                self.reset_factor = 0
                self.reset()
        else:
            self.obj.localOrientation = self._defaults[0]
            self.head.localOrientation = self._defaults[1]

    def update(self):
        game_object_x = self.obj
        game_object_y = self.head
        sensitivity = self.sensitivity * 10
        cap_x = self.cap_x
        lowercapX = cap_x[0] * pi / 180
        uppercapX = cap_x[1] * pi / 180
        cap_y = self.cap_y
        lowercapY = cap_y[0] * pi / 180
        uppercapY = cap_y[1] * pi / 180
        invert = self.invert
        smooth = 1 - (self.smoothing * .99)
        joystick = self.joystick
        raw_values = joystick.axisValues
        if self.stick == 'RS':
            x, y = raw_values[2], raw_values[3]
        elif self.stick == 'LS':
            x, y = raw_values[0], raw_values[1]
        neg_x = -1 if x < 0 else 1
        neg_y = -1 if y < 0 else 1

        threshold = self.threshold
        if -threshold < x < threshold:
            x = 0
        else:
            x = abs(x) ** self.exponent

        if -threshold < y < threshold:
            y = 0
        else:
            y = abs(y) ** self.exponent

        x *= neg_x
        y *= neg_y

        self._x = x = interpolate(self._x, -x if invert[0] else x, smooth)
        self._y = y = interpolate(self._y, -y if invert[1] else y, smooth)
        if self._x == self._y == 0:
            self.done = True
            return

        x *= sensitivity
        y *= sensitivity
        if self.use_cap_x:
            objectRotation = game_object_x.localOrientation.to_euler()

            if objectRotation.z + x > uppercapX:
                x = 0
                objectRotation.z = uppercapX
                game_object_x.localOrientation = objectRotation

            if objectRotation.z + x < lowercapX:
                x = 0
                objectRotation.z = lowercapX
                game_object_x.localOrientation = objectRotation

        game_object_x.applyRotation((0, 0, x), self.local)

        rot_axis = 1 - self.front
        if self.use_cap_y:
            objectRotation = game_object_y.localOrientation.to_euler()

            if objectRotation[rot_axis] + y > uppercapY:
                objectRotation[rot_axis] = uppercapY
                game_object_y.localOrientation = objectRotation
                y = 0

            if objectRotation[rot_axis] + y < lowercapY:
                objectRotation[rot_axis] = lowercapY
                game_object_y.localOrientation = objectRotation
                y = 0

        rot = [0, 0, 0]
        rot[1-self.front] = y
        game_object_y.applyRotation((*rot, ), True)
        self.done = True


class GamepadLook(ULGamepadLook):
    _deprecated = False



from .button import Button, LabelButton
from .cursor import Cursor, remove_custom_cursor
from .image import Image, Icon
from .label import Label
from .layout import FloatLayout
from .layout import Layout
from .layout import RelativeLayout
from .layout import BoxLayout
from .layout import GridLayout
from .widget import Widget
from .canvas import Canvas



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULSetMousePosition(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.screen_x = None
        self.screen_y = None
        self.network = None
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def setup(self, network):
        self.network = network

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        screen_x = self.get_input(self.screen_x)
        screen_y = self.get_input(self.screen_y)
        if is_waiting(screen_x, screen_y):
            return
        self._set_ready()
        self.network.set_mouse_position(screen_x, screen_y)
        self.done = True



from uplogic.data import GlobalDB
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_INVALID
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting


class ULGetGlobalValue(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.data_id = None
        self.key = None
        self.default = None
        self.OUT = ULOutSocket(self, self.get_out)

    def get_out(self):
        socket = self.get_output('val')
        if socket is None:
            data_id = self.get_input(self.data_id)
            key = self.get_input(self.key)
            default = self.get_input(self.default)
            if default is STATUS_INVALID:
                default = None
            if is_waiting(data_id, key, default):
                return STATUS_WAITING
            db = GlobalDB.retrieve(data_id)
            return self.set_output('val', db.get(key, default))
        return socket

    def evaluate(self):
        self._set_ready()



from .buoyancy import ULFlotsam, Flotsam  # noqa
from .buoyancy import ULShip, Ship  # noqa
from .collision import on_collision  # noqa
from .vehicle import FOURWD  # noqa
from .vehicle import FWD  # noqa
from .vehicle import RWD  # noqa
from .vehicle import ULVehicle, Vehicle  # noqa
from .character import ULCharacter, Character  # noqa
from .constraints import create_constraint  # noqa
from .constraints import remove_constraint  # noqa
from .constraints import ULSpring, Spring
from .constraints import ULTrackTo, TrackTo



from .widget import Widget
import gpu
from bge import render


class Layout(Widget):

    def __init__(
        self,
        pos: list = [0., 0.],
        size: list = [100., 100.],
        bg_color: list = [0, 0, 0, 0],
        relative: dict = {},
        border_width: int = 1,
        border_color: list = [0, 0, 0, 0],
        halign: str = 'left',
        valign: str = 'bottom'
    ):
        super().__init__(pos, size, bg_color, relative, halign=halign, valign=valign)
        self.border_width = border_width
        self.border_color = border_color

    @property
    def opacity(self):
        return self.bg_color[3]

    @opacity.setter
    def opacity(self, val):
        self.bg_color[3] = val
        self.border_color[3] = val

    @property
    def border_color(self):
        return self._border_color

    @border_color.setter
    def border_color(self, color):
        self._border_color = list(color)

    @property
    def border_width(self):
        return self._border_width

    @border_width.setter
    def border_width(self, val):
        self._border_width = int(val)

    def draw(self):
        super()._setup_draw()
        gpu.state.line_width_set(self.border_width)
        gpu.state.point_size_set(self.border_width)
        self._shader.uniform_float("color", self.bg_color)
        self._batch.draw(self._shader)
        self._shader.uniform_float("color", self.border_color)
        self._batch_line.draw(self._shader)
        self._batch_points.draw(self._shader)
        super().draw()


class RelativeLayout(Layout):
    @property
    def clipping(self):
        return [
            0,
            render.getWindowWidth(),
            render.getWindowHeight(),
            0
        ]


class FloatLayout(Layout):

    @property
    def pos_abs(self):
        return [0, 0]


class BoxLayout(Layout):
    def __init__(
            self,
            orientation: str = 'horizontal',
            pos: list = [0, 0],
            size: list = [100, 100],
            bg_color: list = (0, 0, 0, 0),
            relative: dict = {},
            border_width: int = 1,
            border_color: list = (0, 0, 0, 0),
            spacing: int = 0,
            halign: str = 'left',
            valign: str = 'bottom'
        ):
        self.orientation = orientation
        self.spacing = spacing
        super().__init__(pos, size, bg_color, relative, border_width, border_color, halign=halign, valign=valign)
        self.use_clipping = False

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, val):
        if self.parent is not val and self.parent:
            self.parent.remove_widget(self)
        if self.use_clipping is None:
            self.use_clipping = val.use_clipping
        self._parent = val
        self.pos = self.pos
        self.size = self.size
        self.arrange()

    def add_widget(self, widget):
        super().add_widget(widget)
        self.arrange()
    
    def remove_widget(self, widget):
        super().remove_widget(widget)
        self.arrange()

    def arrange(self):
        dsize = self._draw_size
        if self.orientation == 'horizontal':
            offset = 0
            for widget in self.children:
                widget.relative['pos'] = False
                widget.pos = [offset, dsize[1] - widget._draw_size[1]]
                offset += widget._draw_size[0] + self.spacing
        if self.orientation == 'vertical':
            offset = dsize[1]
            for widget in self.children:
                offset -= widget._draw_size[1]
                widget.relative['pos'] = False
                widget.pos = [0, offset]
                offset -= self.spacing


class GridLayout(BoxLayout):


    def __init__(
        self,
        orientation: str = 'vertical',
        pos: list = [0, 0],
        size: list = [100, 100],
        bg_color: list = [0, 0, 0, 0],
        relative: dict = {},
        border_width: int = 1,
        border_color: list = [0, 0, 0, 0],
        spacing: int = 0,
        cols: int = 2,
        rows: int = 2,
        halign: str = 'left',
        valign: str = 'bottom'
    ):
        super().__init__(
            orientation,
            pos,
            size,
            bg_color,
            relative,
            border_width,
            border_color,
            spacing,
            halign,
            valign
        )
        self.rows = rows
        self.cols = cols

    def add_widget(self, widget):
        max = self.rows * self.cols
        print(len(self.children), max)
        if len(self.children) < max:
            super().add_widget(widget)
            self.arrange()

    def arrange(self):
        dsize = self._draw_size
        idx = 0
        if self.orientation == 'horizontal':
            row = 0
            offset = 0
            for widget in self.children:
                offset_y = self._draw_size[1] / (self.rows) * row
                widget.relative['pos'] = False
                widget.pos = [offset, dsize[1] - widget._draw_size[1] - offset_y]
                offset += widget._draw_size[0] + self.spacing
                idx += 1
                if idx >= self.cols:
                    idx = 0
                    row += 1
                    offset = 0
        if self.orientation == 'vertical':
            col = 0
            offset = dsize[1]
            for widget in self.children:
                offset_x = self._draw_size[0] / (self.cols - 1) * col
                offset -= widget._draw_size[0]
                widget.relative['pos'] = False
                widget.pos = [offset_x, offset]
                offset -= self.spacing
                idx += 1
                if idx >= self.cols:
                    idx = 0
                    col += 1
                    offset = dsize[1]



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid


class ULReplaceMesh(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.target_game_object = None
        self.new_mesh_name = None
        self.use_display = None
        self.use_physics = None
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        self._set_ready()
        if not condition:
            return
        target = self.get_input(self.target_game_object)
        mesh = self.get_input(self.new_mesh_name)
        display = self.get_input(self.use_display)
        physics = self.get_input(self.use_physics)
        if is_invalid(target):
            return
        if mesh is None:
            return
        if display is None:
            return
        if physics is None:
            return
        target.replaceMesh(mesh, display, physics)
        if physics:
            target.reinstancePhysicsMesh()
        self.done = True



from uplogic.nodes import ULConditionNode
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULBarrier(ULConditionNode):
    consumed: bool
    trigger: float

    def __init__(self):
        ULConditionNode.__init__(self)
        self.condition = None
        self.time = None
        self.consumed = False
        self.trigger = 0

    def evaluate(self):
        condition = self.get_input(self.condition)
        time = self.get_input(self.time)
        if is_waiting(time):
            return

        self._set_ready()

        now = self.network.timeline

        if not not_met(condition):
            if not self.consumed:
                self.consumed = True
                self.trigger = now + time

            if now >= self.trigger:
                self._set_value(True)

        else:
            self._set_value(False)
            self.consumed = False



from uplogic.nodes import ULConditionNode
from uplogic.utils import is_waiting


class ULAndNot(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.condition_a = None
        self.condition_b = None

    def evaluate(self):
        ca = self.get_input(self.condition_a)
        self._set_ready()
        if is_waiting(ca) or not ca:
            self._set_value(False)
            return
        cb = not self.get_input(self.condition_b)
        if is_waiting(cb) or not cb:
            self._set_value(False)
            return
        self._set_value(True)



'''TODO: Documentation
'''

from bge import logic
from uplogic.data.globaldb import GlobalDB
from uplogic.input.vr import ULHeadsetVRWrapper
from uplogic.utils import check_vr_session_status
import aud
import bpy


DISTANCE_MODELS = {
    'EXPONENT': aud.DISTANCE_MODEL_EXPONENT,
    'EXPONENT_CLAMPED': aud.DISTANCE_MODEL_EXPONENT_CLAMPED,
    'INVERSE': aud.DISTANCE_MODEL_INVERSE,
    'INVERSE_CLAMPED': aud.DISTANCE_MODEL_INVERSE_CLAMPED,
    'LINEAR': aud.DISTANCE_MODEL_LINEAR,
    'LINEAR_CLAMPED': aud.DISTANCE_MODEL_LINEAR_CLAMPED,
    'NONE': aud.DISTANCE_MODEL_INVALID
}


def set_lowpass(frequency, system_name='default') -> None:
    """Set the overall volume of a `ULAudioSystem`. All sounds played via this
    system will have their volume multiplied by this value.
    """
    aud_sys = get_audio_system(system_name)
    if aud_sys:
        aud_sys.lowpass = frequency


def set_master_volume(volume, system_name='default') -> None:
    """Set the overall volume of a `ULAudioSystem`. All sounds played via this
    system will have their volume multiplied by this value.
    """
    aud_sys = get_audio_system(system_name)
    if aud_sys:
        aud_sys.volume = volume
        for sound in aud_sys._active_sounds:
            sound.volume = sound.volume


def set_vr_audio(flag, system_name='default') -> None:
    """Set the audio mode for a `ULAudioSystem`. If set to `True`, the system
    will track a VR Headset instead of the active scene camera.
    """
    aud_sys = get_audio_system(system_name)
    if aud_sys:
        aud_sys.use_vr = flag


def stop_all_audio() -> None:
    """Stop every `ULAudioSystem` in this scene.
    """
    for sys in GlobalDB.retrieve('uplogic.audio'):
        sys.shutdown()


class ULAudioSystem(object):
    '''System for managing sounds started using `Sound2D` or `Sound3D`.

    This is usually addressed indirectly through `Sound2D` or `Sound3D` and
    is not intended for manual use.
    '''
    def __init__(self, name: str, mode: str = '3D'):
        self._active_sounds = []
        self.name = name
        self.mode = mode
        self.bounces = 0
        self.volume = 1.0
        self.reverb = False
        self._lowpass = False
        self.device = aud.Device()
        self.device.distance_model = DISTANCE_MODELS[bpy.context.scene.audio_distance_model]
        self.device.speed_of_sound = bpy.context.scene.audio_doppler_speed
        self.device.doppler_factor = bpy.context.scene.audio_doppler_factor
        self._reverb_volumes = []
        self.scene = logic.getCurrentScene()
        self.use_vr = getattr(bpy.data.scenes[self.scene.name], 'use_vr_audio_space', False)
        self.vr_headset = ULHeadsetVRWrapper() if check_vr_session_status() else None
        self.listener = self.vr_headset if self.use_vr else self.scene.active_camera
        self.old_lis_pos = self.listener.worldPosition.copy()
        self.setup(self.scene)
        self.scene.onRemove.append(self.shutdown)

    @property
    def lowpass(self):
        return self._lowpass

    @lowpass.setter
    def lowpass(self, val):
        if val == self._lowpass:
            return
        self._lowpass = val
        for sound in self._active_sounds:
            sound.lowpass = val

    def setup(self, scene=None):
        """Get necessary scene data.
        """
        if scene is None:
            self.scene = logic.getCurrentScene()
        else:
            self.scene = scene
        for obj in self.scene.objects:
            if getattr(obj.blenderObject, 'reverb_volume', False) and not obj.blenderObject.data:
                self._reverb_volumes.append(obj)
        self.reverb = len(self._reverb_volumes) > 0
        GlobalDB.retrieve('uplogic.audio').put(self.name, self)
        self.scene.pre_draw.append(self.update)

    def get_distance_model(self, name):
        return DISTANCE_MODELS.get(name, aud.DISTANCE_MODEL_INVERSE_CLAMPED)

    def compute_listener_velocity(self, listener):
        """Compare positions of the listener to calculate velocity.
        """
        wpos = listener.worldPosition.copy()
        olp = self.old_lis_pos
        vel = (
            (wpos.x - olp.x) * 50,
            (wpos.y - olp.y) * 50,
            (wpos.z - olp.z) * 50
        )
        self.old_lis_pos = wpos
        return vel

    def update(self):
        """This is called each frame.
        """
        if self.mode == '3D':
            scene = logic.getCurrentScene()
            if scene is not self.scene:
                self.setup(scene)
            listener = self.vr_headset if self.use_vr else scene.active_camera
            self.reverb = False
            same_cam = listener is self.listener
            if not self.use_vr:
                self.listener = listener
            if not self._active_sounds:
                return  # do not update if no sound has been installed
            # update the listener data
            cpos = listener.worldPosition
            distances = {}
            if self._reverb_volumes:
                for obj in self._reverb_volumes:
                    dist = (obj.worldPosition - cpos).length
                    if dist > 50:
                        continue
                    else:
                        distances[dist] = obj
                if distances:
                    min_dist = distances[min(distances.keys())]
                    obj = min_dist
                    ob = obj.blenderObject
                    r = ob.empty_display_size
                    wpos = obj.worldPosition
                    sca = ob.scale
                    in_range = (
                        wpos.x - r*sca.x < cpos.x < wpos.x + r*sca.x and
                        wpos.y - r*sca.y < cpos.y < wpos.y + r*sca.y and
                        wpos.z - r*sca.z < cpos.z < wpos.z + r*sca.z
                    )
                    if in_range:
                        self.reverb = True
                        self.bounces = ob.reverb_samples
            listener_vel = (0, 0, 0) if self.use_vr or not same_cam else self.compute_listener_velocity(listener)
            dev = self.device
            dev.listener_location = cpos
            dev.listener_orientation = listener.worldOrientation.to_quaternion()
            dev.listener_velocity = listener_vel
        for s in self._active_sounds:
            s.update()

    def add(self, sound):
        '''Add a `ULSound` to this audio system.'''
        self._active_sounds.append(sound)

    def remove(self, sound):
        '''Remove a `ULSound` from this audio system.'''
        self._active_sounds.remove(sound)

    def shutdown(self, a=None):
        '''Stop and remove this audio system. This will stop all sounds playing
        on this system.'''
        self.device.stopAll()
        self.scene.pre_draw.remove(self.update)
        GlobalDB.retrieve('uplogic.audio').remove(self.name)


def get_audio_system(system_name: str = 'default', mode: str = '3D') -> ULAudioSystem:
    '''Get or create a `ULAudioSystem` with the given name.

    :param `system_name`: Look for this name.

    :returns: `ULAudioSystem`, new system is created if none is found.
    '''
    scene = logic.getCurrentScene()
    aud_systems = GlobalDB.retrieve('uplogic.audio')
    if aud_systems.check(system_name):
        aud_sys = aud_systems.get(system_name)
    else:
        aud_sys = ULAudioSystem(system_name, mode)
    if aud_sys.update not in scene.pre_draw:
        scene.pre_draw.append(aud_sys.update)
    return aud_sys


from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import debug
from uplogic.utils import not_met
import bpy
import json
import os


class ULClearVariables(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.file_name = None
        self.path = ''
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def write_to_json(self, path):
        data = None
        if not path.endswith('.json'):
            path = path + f'{self.get_input(self.file_name)}.json'
        if os.path.isfile(path):
            data = {}
            f = open(path, 'w')
            json.dump(data, f, indent=2)
        else:
            debug('File does not exist - creating...')
            f = open(path, 'w')
            data = {}
            json.dump(data, f, indent=2)
        f.close()

    def get_custom_path(self, path):
        if not path.endswith('/') and not path.endswith('json'):
            path = path + '/'
        return path

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        self._set_ready()
        cust_path = self.get_custom_path(self.path)

        path = (
            bpy.path.abspath('//Data/')
            if self.path == ''
            else bpy.path.abspath(cust_path)
        )
        os.makedirs(path, exist_ok=True)

        self.write_to_json(path)
        self.done = True



from uplogic.nodes import ULActionNode
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULSendMessage(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.from_obj = None
        self.to_obj = None
        self.subject = None
        self.body = None

    def evaluate(self):
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_ready()
            self._set_value(False)
            return
        from_obj = self.get_input(self.from_obj)
        to_obj = self.get_input(self.to_obj)
        subject = self.get_input(self.subject)
        body = self.get_input(self.body)
        if is_waiting(from_obj, to_obj, subject, body):
            return
        self._set_ready()
        if body and to_obj:
            from_obj.sendMessage(subject, body=body, to=to_obj)
        elif body:
            from_obj.sendMessage(subject, body=body)
        elif to_obj:
            from_obj.sendMessage(subject, to=to_obj)
        else:
            from_obj.sendMessage(subject)
        self._set_value(True)



from bge import logic
from uplogic.audio import Sound3D
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import not_met


class ULStartSpeaker(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.occlusion = None
        self.transition = None
        self.cutoff = None
        self.speaker = None
        self.done = None
        self.on_finish = False
        self._clear_sound = 1
        self._sustained = 1
        self._handle = None
        self.DONE = ULOutSocket(self, self.get_done)
        self.ON_FINISH = ULOutSocket(self, self.get_on_finish)
        self.HANDLE = ULOutSocket(self, self.get_handle)

    def get_handle(self):
        return self._handle

    def get_on_finish(self):
        if not self._handle:
            return False
        if self._handle.finished:
            self._handle = None
            return True
        return False

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        self.on_finish = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_ready()
            return
        speaker = self.get_input(self.speaker)
        transition = self.get_input(self.transition)
        occlusion = self.get_input(self.occlusion)
        cutoff = self.get_input(self.cutoff)
        loop_count = self.get_input(self.loop_count)
        bl_speaker = speaker.blenderObject.data
        file = bl_speaker.sound.filepath
        if not file:
            return
        volume = bl_speaker.volume
        pitch = bl_speaker.pitch * logic.getTimeScale()
        attenuation = bl_speaker.attenuation
        distance_ref = bl_speaker.distance_reference
        cone_inner = bl_speaker.cone_angle_inner
        cone_outer = bl_speaker.cone_angle_outer
        cone_outer_volume = bl_speaker.cone_volume_outer
        self._set_ready()

        if is_invalid(file):
            return
        self._handle = Sound3D(
            file,
            speaker,
            'ln_audio_system',
            occlusion,
            transition,
            cutoff,
            volume,
            pitch,
            attenuation,
            distance_ref,
            [cone_inner, cone_outer],
            cone_outer_volume,
            loop_count
        )
        self.done = True



from .shader import Filter2D
from bge import logic


glsl = """
#version 120
#pragma optionNV(fastmath on)
#pragma optionNV(fastprecision on)
#pragma optionNV(inline all)
#pragma optionNV(unroll all)
#pragma optionNV(ifcvt none)
#pragma optionNV(strict on)

uniform float sharpness;

uniform sampler2D bgl_RenderedTexture;
uniform float bgl_RenderedTextureWidth;
uniform float bgl_RenderedTextureHeight;

in vec4 bgl_TexCoord;

out vec4 fragColor;

void main(void)
{
    float width = 1.0 /  bgl_RenderedTextureWidth; 
    float height = 1.0 / bgl_RenderedTextureHeight;

    vec4 Total = vec4(0.0);

    vec2 Scoord = bgl_TexCoord.xy;
    vec4 Msample = texture(bgl_RenderedTexture, Scoord);

    float Hoffs = 1.5 * width;
    float Voffs = 1.5 * height;

    vec2 UV1 = bgl_TexCoord.xy + vec2(Hoffs,Voffs);  
    Total += texture(bgl_RenderedTexture, UV1);

    vec2 UV2 = bgl_TexCoord.xy + vec2(-Hoffs,Voffs);  
    Total += texture(bgl_RenderedTexture, UV2);

    vec2 UV3 = bgl_TexCoord.xy + vec2(-Hoffs,-Voffs);  
    Total += texture(bgl_RenderedTexture, UV3);

    vec2 UV4 = bgl_TexCoord.xy + vec2(Hoffs,-Voffs);  
    Total += texture(bgl_RenderedTexture, UV4);

    Total *= .15;

    Hoffs = 2.5 * width;
    Voffs = 2.5 * height;

    vec2 UV5 = bgl_TexCoord.xy + vec2(Hoffs,0.0);  
    Total += texture(bgl_RenderedTexture, UV5) *.1;

    vec2 UV6 = bgl_TexCoord.xy + vec2(0.0,Voffs);  
    Total += texture(bgl_RenderedTexture, UV6) *.1;

    vec2 UV7 = bgl_TexCoord.xy + vec2(-Hoffs,0.0);  
    Total += texture(bgl_RenderedTexture, UV7) *.1;

    vec2 UV8 = bgl_TexCoord.xy + vec2(0.0,-Voffs);  
    Total += texture(bgl_RenderedTexture, UV8) *.1;

    vec4 Final = (1 + sharpness) * Msample - Total * sharpness;

    fragColor.xyz = Final.xyz;
    fragColor.a = 1.0;
}
"""


class Sharpen(Filter2D):

    def __init__(self, sharpness=0.0, idx: int = None) -> None:
        self.settings = {'sharpness': float(sharpness)}
        super().__init__(glsl, idx, {'sharpness': self.settings})

    @property
    def sharpness(self):
        return self.settings['sharpness']

    @sharpness.setter
    def sharpness(self, val):
        self.settings['sharpness'] = val



from uplogic.audio.audiosystem import get_audio_system
from .sound import Sound2D
from uuid import uuid4
from uplogic.utils import debug, interpolate
from uplogic.events import schedule_callback


class ULMusicEffect():
    '''Base class for sounds played as music.'''
    def __init__(
        self
    ):
        self._volume = 1.0
        self._pitch = 1.0
        self._fade_event = None

    @property
    def volume(self):
        '''Volume of this component.'''
        return self._volume

    @volume.setter
    def volume(self, val):
        self._volume = val

    def fade_out(self, factor=.02):
        '''Fade out this sound using linear interpolation.

        :param `factor`: Speed of interpolation.'''
        self.volume = interpolate(self.volume, 0, factor)
        if self._fade_event:
            self._fade_event.cancel()
            self._fade_event = None
        if self.volume > 0:
            self._fade_event = schedule_callback(self.fade_out, arg=factor)

    def fade_in(self, factor=.01):
        '''Fade in this sound using linear interpolation.

        :param `factor`: Speed of interpolation.'''
        self.volume = interpolate(self.volume, 1, factor)
        if self._fade_event:
            self._fade_event.cancel()
            self._fade_event = None
        if self.volume < 1:
            self._fade_event = schedule_callback(self.fade_in, arg=factor)


class ULMusic(ULMusicEffect):
    '''Management class for controlling multiple music tracks.

    :param `name`: Name of this music.'''
    _deprecated = True

    def __init__(
        self,
        name: str = '',
        audio_system: str = 'music'
    ):
        if self._deprecated:
            debug('ULMusic class will be renamed to "Music" in future releases!')
        super().__init__()
        self.name = name if name else uuid4()
        self.audio_system = get_audio_system(audio_system, '2D')
        self.tracks: list[ULMusicTrack] = []

    @property
    def position(self):
        '''Playback position of the first track of the music (and consequently
        all other tracks).'''
        if not self.tracks:
            return 0.0
        return self.tracks[0].sound.position

    @position.setter
    def position(self, val):
        for track in self.tracks:
            track.position = val

    @property
    def volume(self):
        '''Master volume of this music.'''
        return self._volume

    @volume.setter
    def volume(self, val):
        self._volume = val
        for track in self.tracks:
            track.sound.volume = track.volume * val
        if self._fade_event:
            self._fade_event.cancel()
            self._fade_event = None

    @property
    def pitch(self):
        '''Master pitch of this music.'''
        return self._volume

    @pitch.setter
    def pitch(self, val):
        self._pitch = val
        for track in self.tracks:
            track.sound.pitch = track.pitch * val
        if self._fade_event:
            self._fade_event.cancel()
            self._fade_event = None

    def add_track(
        self,
        sound: str or Sound2D,
        track_name: str = ''
    ):
        '''Add a track to this music. A track is typically one instrument or
        effect.
        
        :param `sound`: Path to the sound file or `Sound2D` instance.
        :param `name`: Name of this track (e.g. "Drums")'''
        if not track_name:
            track_name = uuid4()
        track = ULMusicTrack(self, sound, track_name)
        self.tracks.append(track)
        return track

    def remove_track(
        self,
        track: int or str = 0,
    ):
        '''Remove a track from this music.
        
        :param `track`: Index or name of the track to be removed.'''
        if isinstance(track, str):
            for t in self.tracks:
                if t.name == track:
                    t.remove()
                    return
            debug(f'Track "{track}" not found!')
        else:
            self.tracks[track].remove()

    def get_track(self, name):
        '''Get track by name.
        
        :param `name`: Name of the track.'''
        for track in self.tracks:
            if track.name == name:
                return track

    def pause(self):
        '''Pause this music (all tracks).'''
        for track in self.tracks:
            track.sound.pause()

    def resume(self):
        '''Rasume this music (all tracks).'''
        for track in self.tracks:
            track.sound.resume()

    def stop(self):
        '''Stop this music (all tracks).'''
        for track in self.tracks:
            track.sound.stop()
        self.position = 0.0


class Music(ULMusic):
    _deprecated = False


class ULMusicTrack(ULMusicEffect):
    '''Track to be played on a `ULMusic` instance.
    
    :param `music`: The music object this track will be played on.
    :param `sound`: Path to the soundfile of this track.
    :param `name`: Name of this track (e.g. "Drums).'''
    def __init__(
        self,
        music: Music,
        sound: str or Sound2D,
        name: str
    ):
        super().__init__()
        self.music = music
        self.name = name
        sound = Sound2D(
            sound,
            aud_sys=self.music.audio_system
        ) if isinstance(sound, str) else sound
        sound.position = music.position
        self.sound: Sound2D = sound

    @property
    def position(self):
        '''Playback position of this track.'''
        return self.sound.position

    @position.setter
    def position(self, val):
        self.sound.position = val

    @property
    def volume(self):
        '''Volume of this track.'''
        return self._volume

    @volume.setter
    def volume(self, val):
        self._volume = val
        self.sound.volume = val * self.music.volume
        if self._fade_event:
            self._fade_event.cancel()
            self._fade_event = None

    @property
    def pitch(self):
        '''Pitch of this track.'''
        return self._pitch

    @pitch.setter
    def pitch(self, val):
        self._pitch = 1.0
        self.sound.pitch = val * self.music.pitch

    def remove(self):
        '''Stop and remove this track from its `ULMusic` object.'''
        self.sound.stop()
        self.music.tracks.remove(self)



from .shader import Filter2D
from bge import logic, render
from mathutils import Vector


# glsl = """
# /* Original Code by existical @shadertoy: https://www.shadertoy.com/view/Xltfzj
# Modified by Iza Zed for UPBGE
# */

# uniform sampler2D bgl_RenderedTexture;
# uniform sampler2D bgl_DepthTexture;
# uniform float bgl_RenderedTextureWidth;
# uniform float bgl_RenderedTextureHeight;
# vec2 resolution = vec2(bgl_RenderedTextureWidth, bgl_RenderedTextureHeight);

# out vec4 fragColor;
# in vec4 bgl_TexCoord;

# vec2 texcoord = bgl_TexCoord.xy;

# uniform float samples;
# uniform float power;
# uniform float distance;
# uniform float fstop;

# uniform float znear; // = .1; //camera clipping start
# uniform float zfar; // = 500; //camera clipping end

# float linearize(float depth)
# {
#     return -zfar * znear / (depth * (zfar - znear) - zfar);
# }

# void main()
# {
#     float Pi = 6.28318530718;

#     float quality = 3.0;

#     vec2 uv = texcoord;
#     vec4 dcolor = clamp(texture(bgl_DepthTexture, uv), 0, .4);
#     dcolor += clamp(texture(bgl_DepthTexture, uv), .0, .4);

# //    for( float d=0.0; d<Pi; d+=Pi/samples)
# //    {
# //        for(float i = 1.0 / quality; i <= 1.0; i += 1.0 / quality)
# //        {
# //            dcolor = texture(bgl_DepthTexture, uv + vec2(cos(d),sin(d)) * vec2(.02)) * i;
# //        }
# //    }
# //    // dcolor /= quality * samples - 15.0;

#     float depth = linearize(dcolor.x);
#     fragColor = vec4(depth);
#     return;
#     depth = abs(distance - depth);

#     float dpower = (1 - depth) * power;

#     vec2 radius = dpower/resolution.xy * (depth * fstop);
#     radius.x = clamp(radius.x, 0.0, dpower / bgl_RenderedTextureWidth);
#     radius.y = clamp(radius.y, 0.0, dpower / bgl_RenderedTextureHeight);

#     vec4 color = texture(bgl_RenderedTexture, uv);

#     // Blur calculations
#     for( float d=0.0; d<Pi; d+=Pi/samples)
#     {
#         for(float i=1.0/quality; i<=1.0; i+=1.0/quality)
#         {
#             color += texture(bgl_RenderedTexture, uv+vec2(cos(d),sin(d)) * radius * i);
#         }
#     }
#     color /= quality * samples - 15.0;
#     fragColor = color;
# }"""


glsl = """
uniform sampler2D bgl_RenderedTexture;
uniform sampler2D bgl_DepthTexture;
uniform float bgl_RenderedTextureWidth;
uniform float bgl_RenderedTextureHeight;

#define PI  3.14159265

float width = bgl_RenderedTextureWidth; //texture width
float height = bgl_RenderedTextureHeight; //texture height

in vec4 bgl_TexCoord;
vec2 texcoord = bgl_TexCoord.xy;

vec2 texel = vec2(1.0/width,1.0/height);

uniform float distance;  //external focal point value, but you may use autofocus option below
//float distance = 1;

//------------------------------------------
//user variables

uniform int samples; //samples on the first ring
int rings = 5; //ring count

uniform bool autofocus; //use autofocus in shader? disable if you use external distance value
vec2 focus = vec2(0.5,0.5); // autofocus point on screen (0.0,0.0 - left lower corner, 1.0,1.0 - upper right)
float range = 2.0; //focal range
uniform float fstop;
uniform float power; //clamp value of max blur

float threshold = 0.5; //highlight threshold;
float gain = 10.0; //highlight gain;

float bias = 0.4; //bokeh edge bias
float fringe = 0.5; //bokeh chromatic aberration/fringing

bool noise = true; //use noise instead of pattern for sample dithering
float namount = 0.0001; //dither amount

bool depthblur = true; //blur the depth buffer?
float dbsize = 2.0; //depthblursize

/*
next part is experimental
not looking good with small sample and ring count
looks okay starting from samples = 4, rings = 4
*/

bool pentagon = false; //use pentagon as bokeh shape?
float feather = 0.4; //pentagon shape feather

//------------------------------------------


float penta(vec2 coords) //pentagonal shape
{
	float scale = float(rings) - 1.3;
	vec4  HS0 = vec4( 1.0,         0.0,         0.0,  1.0);
	vec4  HS1 = vec4( 0.309016994, 0.951056516, 0.0,  1.0);
	vec4  HS2 = vec4(-0.809016994, 0.587785252, 0.0,  1.0);
	vec4  HS3 = vec4(-0.809016994,-0.587785252, 0.0,  1.0);
	vec4  HS4 = vec4( 0.309016994,-0.951056516, 0.0,  1.0);
	vec4  HS5 = vec4( 0.0        ,0.0         , 1.0,  1.0);

	vec4  one = vec4( 1.0 );

	vec4 P = vec4((coords),vec2(scale, scale)); 

	vec4 dist = vec4(0.0);
	float inorout = -4.0;

	dist.x = dot( P, HS0 );
	dist.y = dot( P, HS1 );
	dist.z = dot( P, HS2 );
	dist.w = dot( P, HS3 );

	dist = smoothstep( -feather, feather, dist );

	inorout += dot( dist, one );

	dist.x = dot( P, HS4 );
	dist.y = HS5.w - abs( P.z );

	dist = smoothstep( -feather, feather, dist );
	inorout += dist.x;

	return clamp( inorout, 0.0, 1.0 );
}

float bdepth(vec2 coords) //blurring depth
{
	float d = 0.0;
	float kernel[9];
	vec2 offset[9];

	vec2 wh = vec2(texel.x, texel.y) * dbsize;

	offset[0] = vec2(-wh.x,-wh.y);
	offset[1] = vec2( 0.0, -wh.y);
	offset[2] = vec2( wh.x -wh.y);

	offset[3] = vec2(-wh.x,  0.0);
	offset[4] = vec2( 0.0,   0.0);
	offset[5] = vec2( wh.x,  0.0);

	offset[6] = vec2(-wh.x, wh.y);
	offset[7] = vec2( 0.0,  wh.y);
	offset[8] = vec2( wh.x, wh.y);

	kernel[0] = 1.0/16.0;   kernel[1] = 2.0/16.0;   kernel[2] = 1.0/16.0;
	kernel[3] = 2.0/16.0;   kernel[4] = 4.0/16.0;   kernel[5] = 2.0/16.0;
	kernel[6] = 1.0/16.0;   kernel[7] = 2.0/16.0;   kernel[8] = 1.0/16.0;


	for( int i=0; i<9; i++ )
	{
		float tmp = texture2D(bgl_DepthTexture, coords + offset[i]).r;
		d += tmp * kernel[i];
	}

	return d;
}


vec3 color(vec2 coords,float blur) //processing the sample
{
	vec3 col = vec3(0.0);

	col.r = texture2D(bgl_RenderedTexture,coords + vec2(0.0,1.0)*texel*fringe*blur).r;
	col.g = texture2D(bgl_RenderedTexture,coords + vec2(-0.866,-0.5)*texel*fringe*blur).g;
	col.b = texture2D(bgl_RenderedTexture,coords + vec2(0.866,-0.5)*texel*fringe*blur).b;

	vec3 lumcoeff = vec3(0.299,0.587,0.114);
	float lum = dot(col.rgb, lumcoeff);
	float thresh = max((lum-threshold)*gain, 0.0);
	return col+mix(vec3(0.0),col,thresh*blur);
}

vec2 rand(in vec2 coord) //generating noise/pattern texture for dithering
{
	float noiseX = ((fract(1.0-coord.s*(width/2.0))*0.25)+(fract(coord.t*(height/2.0))*0.75))*2.0-1.0;
	float noiseY = ((fract(1.0-coord.s*(width/2.0))*0.75)+(fract(coord.t*(height/2.0))*0.25))*2.0-1.0;

	if (noise)
	{
		noiseX = clamp(fract(sin(dot(coord ,vec2(12.9898,78.233))) * 43758.5453),0.0,1.0)*2.0-1.0;
		noiseY = clamp(fract(sin(dot(coord ,vec2(12.9898,78.233)*2.0)) * 43758.5453),0.0,1.0)*2.0-1.0;
	}
	return vec2(noiseX,noiseY);
}

void main() 
{

	float depth = texture2D(bgl_DepthTexture, texcoord).x;
	float blur = 0.0;

	if (depthblur)
	{
		depth = bdepth(texcoord);
	}

	blur = clamp((abs(depth - distance)/fstop)*100.0,-power,power);

	if (autofocus)
	{
		float fDepth = texture2D(bgl_DepthTexture, focus).x;
		blur = clamp((abs(depth - fDepth)/fstop)*100.0,-power,power);
	}

	vec2 noise = rand(texcoord)*namount*blur;

	float w = (1.0/width)*blur+noise.x;
	float h = (1.0/height)*blur+noise.y;

	vec3 col = texture2D(bgl_RenderedTexture, texcoord).rgb;
	float s = 1.0;

	int ringsamples;

	for (int i = 1; i <= rings; i += 1)
	{   
		ringsamples = i * samples;
		 
		for (int j = 0 ; j < ringsamples ; j += 1)   
		{
			float step = PI*2.0 / float(ringsamples);
			float pw = (cos(float(j)*step)*float(i));
			float ph = (sin(float(j)*step)*float(i));
			float p = 1.0;
			if (pentagon)
			{ 
			p = penta(vec2(pw,ph));
			}
			col += color(texcoord + vec2(pw*w,ph*h),blur)*mix(1.0,(float(i))/(float(rings)),bias)*p;  
			s += 1.0*mix(1.0,(float(i))/(float(rings)),bias)*p;   
		}
	}

	col /= s;   

	gl_FragColor.rgb = col;
	gl_FragColor.a = 1.0;
}
"""


class DoF(Filter2D):

        
    def __init__(self, distance=1.0, autofocus=False, power=1.0, fstop=1, samples=16, idx: int = None) -> None:
        cam = logic.getCurrentScene().active_camera
        self.settings = {
            'distance': float(distance),
            'power': float(power),
            'fstop': float(fstop),
            'samples': int(samples),
            'autofocus': bool(autofocus)
            # 'znear': cam.near,
            # 'zfar': cam.far
        }
        super().__init__(glsl, idx, {
            'distance': self.settings,
            'power': self.settings,
            'fstop': self.settings,
            'samples': self.settings,
            'autofocus': self.settings
            # 'znear': self.settings,
            # 'zfar': self.settings
        })

    def update(self):
        super().update()
        cam = logic.getCurrentScene().active_camera
        self.settings['znear'] = cam.near
        self.settings['zfar'] = cam.far

    @property
    def distance(self):
        return self.settings['distance']

    @distance.setter
    def distance(self, val):
        self.settings['distance'] = float(val)

    @property
    def autofocus(self):
        return self.settings['autofocus']

    @autofocus.setter
    def autofocus(self, val):
        self.settings['autofocus'] = bool(val)

    @property
    def fstop(self):
        return self.settings['fstop']

    @fstop.setter
    def fstop(self, val):
        self.settings['fstop'] = float(val)

    @property
    def power(self):
        return self.settings['power']

    @power.setter
    def power(self, val):
        self.settings['power'] = val

    @property
    def samples(self):
        return self.settings['samples']

    @samples.setter
    def samples(self, val):
        self.settings['samples'] = int(val)



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING, is_invalid
from uplogic.utils import is_waiting
from uplogic.utils import get_local
from uplogic.utils import not_met


class ULActionRotateTo(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.moving_object = None
        self.target_point = None
        self.speed = None
        self._done = False
        self.rot_axis = 2
        self.front_axis = 0
        self.OUT = ULOutSocket(self, self._get_done)

    def _get_done(self):
        return self._done

    def evaluate(self):
        self._done = False
        self._set_value(False)
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        moving_object = self.get_input(self.moving_object)
        target_point = self.get_input(self.target_point)
        speed = self.get_input(self.speed)
        target_point = getattr(target_point, 'worldPosition', target_point)
        rot_axis = self.get_input(self.rot_axis)
        front_axis = self.get_input(self.front_axis)
        if is_invalid(moving_object):
            return
        if is_waiting(target_point, speed, rot_axis, front_axis):
            return
        self._set_ready()
        aim = get_local(moving_object, target_point)
        if front_axis > 2:
            speed = -speed
            front_axis -= 3
        if rot_axis == front_axis:
            return
        if rot_axis == 0:
            aim = aim.yz.normalized()
            rot = -aim.x if front_axis == 2 else aim.y
            moving_object.applyRotation((rot * speed, 0, 0), True)
        elif rot_axis == 1:
            aim = aim.xz.normalized()
            rot = aim.x if front_axis == 2 else -aim.y
            moving_object.applyRotation((0, rot * speed, 0), True)
        elif rot_axis == 2:
            aim = aim.xy.normalized()
            rot = -aim.x if front_axis == 1 else aim.y
            moving_object.applyRotation((0, 0, rot * speed), True)
        self._done = True



from bge import events
from uplogic.nodes import ULConditionNode
from uplogic.input import mouse_moved


class ULMouseMoved(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.pulse = False

    def evaluate(self):
        self._set_ready()
        if self.pulse:
            self._set_value(
                mouse_moved()
            )
        else:
            self._set_value(
                mouse_moved(tap=True)
            )



from math import pi
from bge import logic
from bge import events
from bge import render
from bge.types import KX_GameObject as GameObject
from mathutils import Vector
from uplogic.utils import interpolate
from uplogic.events import schedule_callback
from uplogic.utils import debug


MOUSE_EVENTS = logic.mouse.inputs
'''Reference to `bge.logic.mouse.inputs`
'''

LMB = events.LEFTMOUSE
RMB = events.RIGHTMOUSE
MMB = events.MIDDLEMOUSE

MOUSE_BUTTONS = {
    'LMB': LMB,
    'RMB': RMB,
    'MMB': MMB
}


def set_mouse_position(x: int, y: int, absolute: bool = False) -> None:
    if absolute:
        render.setMousePosition(x, y)
        return
    render.setMousePosition(
        int(x * render.getWindowWidth()),
        int(y * render.getWindowHeight())
    )


def get_mouse_position(absolute: bool = False) -> Vector:
    pos = logic.mouse.position
    if absolute:
        return Vector((
            int(pos[0] * render.getWindowWidth()),
            int(pos[1] * render.getWindowHeight())
        ))
    return Vector(pos)


def mouse_moved(tap: bool = False) -> bool:
    '''Detect mouse movement.

    :param tap: Only use the first consecutive `True` output

    :returns: boolean
    '''
    if tap:
        return (
            MOUSE_EVENTS[events.MOUSEX].activated or
            MOUSE_EVENTS[events.MOUSEY].activated
        )
    else:
        return (
            MOUSE_EVENTS[events.MOUSEX].active or
            MOUSE_EVENTS[events.MOUSEY].active or
            MOUSE_EVENTS[events.MOUSEX].activated or
            MOUSE_EVENTS[events.MOUSEY].activated
        )


def mouse_tap(button=events.LEFTMOUSE) -> bool:
    '''Detect mouse button tap.

    :param button: can be either `LMB`, `RMB` or `MMB` from `uplogic.input`

    :returns: boolean
    '''
    button = MOUSE_BUTTONS.get(button, button)
    return (
        MOUSE_EVENTS[button].activated or
        MOUSE_EVENTS[button].activated
    )


def mouse_down(button=events.LEFTMOUSE) -> bool:
    '''Detect mouse button held down.

    :param button: can be either `LMB`, `RMB` or `MMB` from `uplogic.input`

    :returns: boolean
    '''
    button = MOUSE_BUTTONS.get(button, button)
    return (
        MOUSE_EVENTS[button].active or
        MOUSE_EVENTS[button].activated or
        MOUSE_EVENTS[button].active or
        MOUSE_EVENTS[button].activated
    )


def mouse_press(button=events.LEFTMOUSE, down=False) -> bool:
    '''Detect mouse button tap or held down.

    :param button: can be either `LMB`, `RMB` or `MMB` from `uplogic.input`

    :returns: boolean
    '''
    button = MOUSE_BUTTONS.get(button, button)
    return (
        MOUSE_EVENTS[button].active or
        MOUSE_EVENTS[button].activated or
        MOUSE_EVENTS[button].active or
        MOUSE_EVENTS[button].activated
    ) if down else (
        MOUSE_EVENTS[button].activated or
        MOUSE_EVENTS[button].activated
    )


def mouse_up(button=events.LEFTMOUSE) -> bool:
    '''Detect mouse button released.

    :param button: can be either `LMB`, `RMB` or `MMB` from `uplogic.input`

    :returns: boolean
    '''
    button = MOUSE_BUTTONS.get(button, button)
    return (
        MOUSE_EVENTS[button].released
    )


def mouse_wheel(tap: bool = False) -> int:
    '''Detect mouse wheel activity.

    :param tap: Only use the first consecutive `True` output

    :returns: -1 if wheel down, 0 if idle, 1 if wheel up
    '''
    if tap:
        return (
            MOUSE_EVENTS[events.WHEELUPMOUSE].activated -
            MOUSE_EVENTS[events.WHEELDOWNMOUSE].activated
        )
    else:
        return (
            (
                MOUSE_EVENTS[events.WHEELUPMOUSE].activated or
                MOUSE_EVENTS[events.WHEELUPMOUSE].active
            ) - (
                MOUSE_EVENTS[events.WHEELDOWNMOUSE].activated or
                MOUSE_EVENTS[events.WHEELDOWNMOUSE].active
            )
        )


class ULMouse():
    """Mouse Wrapper for accessing mouse data."""

    _deprecated = True

    def __init__(self) -> None:
        if self._deprecated:
            debug('ULMouse class will be renamed to "Mouse" in future releases!')
        self._position = get_mouse_position()
        """Staggered updated mouse position for pos difference calculation."""
        self.movement = (0, 0)
        """Movement of the mouse as a tuple `(x, y)`."""
        self.active = True

    @property
    def active(self):
        """Tracking state of this component."""
        return self.update in logic.getCurrentScene().pre_draw

    @active.setter
    def active(self, val):
        pre_draw = logic.getCurrentScene().pre_draw
        if val and self.update not in pre_draw:
            pre_draw.append(self.update)
        elif not val and self.update in pre_draw:
            pre_draw.remove(self.update)

    @property
    def position(self):
        """Position of the mouse as a tuple of `(x, y)` ranging from 0-1 on both
        axis."""
        return get_mouse_position()

    @position.setter
    def position(self, val):
        render.setMousePosition(
            int(val[0] * render.getWindowWidth()),
            int(val[1] * render.getWindowHeight())
        )
        self._position = get_mouse_position()

    @property
    def moved(self):
        """`True` if the mouse is moved, `False` if idle."""
        return mouse_moved()

    @moved.setter
    def moved(self, val):
        print('ULMouse.moved is read-only!')

    @property
    def wheel(self):
        """Mouse wheel difference.
        
        -1 if scolled down, 0 if idle, 1 if scrolled up."""
        return mouse_wheel()

    @wheel.setter
    def wheel(self, val):
        print('ULMouse.wheel is read-only!')

    def update(self) -> None:
        """This is executed each frame if component is active."""
        old_pos = self._position
        new_pos = self.position
        self.movement = (
            new_pos[0] - old_pos[0],
            new_pos[1] - old_pos[1]
        )
        self._position = new_pos

    def button_down(self, button: str = 'LMB'):
        """Check if a button on the mouse is held down.
        
        :param `button`: The button to check for; `str` of [`'LMB'`, `'MMB'`,
        `'RMB'`]"""
        return mouse_down(MOUSE_BUTTONS[button])

    def button_up(self, button: str = 'LMB'):
        """Check if a button on the mouse is released.
        
        :param `button`: The button to check for; `str` of [`'LMB'`, `'MMB'`,
        `'RMB'`]"""
        return mouse_up(MOUSE_BUTTONS[button])

    def button_tap(self, button: str = 'LMB'):
        """Check if a button on the mouse is pressed once.
        
        :param `button`: The button to check for; `str` of [`'LMB'`, `'MMB'`,
        `'RMB'`]"""
        return mouse_tap(MOUSE_BUTTONS[button])


class Mouse(ULMouse):
    _deprecated = False

MOUSE = Mouse()


class ULMouseLook():
    """Automatically track the mouse movement and translate it to a rotate a
    body and optionally a head.

    This component can be activated/deactivated at any time to keep performance
    up.

    :param `obj`: Main object to rotate around the object's Z axis.
    :param `head`: Head object to rotate around the object's X/Y axis.
    :param `sensitivity`: Translation factor of mouse movement to rotation.
    :param `use_cap_x`: Whether to use capping on the mouse X movement (Z axis
    rotation).
    :param `cap_x`: Minimum and Maximum amount of rotation on the Z axis.
    :param `use_cap_y`: Whether to use capping on the mouse Y movement (X/Y axis
    rotation).
    :param `cap_y`: Minimum and Maximum amount of rotation on the X/Y axis.
    :param `invert`: Whether to use inverted values for mous X/Y movement.
    :param `smoothing`: Amount of movement smoothing.
    :param `local`: Whether to use local transform for the body object.
    :param `front`: Front axis (traditionally in blender, Y is front).
    :param `active`: Whether to start this component in active or inactive mode
    (can be changed later).
    """

    _deprecated = True

    def __init__(
        self,
        obj: GameObject,
        head: GameObject = None,
        sensitivity: float = 1.0,
        use_cap_x: bool = False,
        cap_x: tuple = (0, 0),
        use_cap_y: bool = False,
        cap_y: tuple = (-89, 89),
        invert: tuple = (False, True),
        smoothing: float = 0.0,
        local: bool = True,
        front: int = 1,
        active: bool = True
    ) -> None:
        if self._deprecated:
            debug('ULMouseLook class will be renamed to "MouseLook" in future releases!')
        self.obj = obj
        self.head = head if head else obj
        self._defaults = [
            obj.localOrientation.copy(),
            head.localOrientation.copy()
            if head else
            obj.localOrientation.copy()
        ]
        self.sensitivity = sensitivity
        self.use_cap_x = use_cap_x
        self.cap_x = cap_x
        self.use_cap_y = use_cap_y
        self.cap_y = cap_y
        self.invert = invert
        self.smoothing = smoothing
        self.initialized = False
        self.front = front
        self._x = 0
        self._y = 0
        self.local = local
        self.axis_lock = [False, False]
        self.reset_factor = 0
        self.active = active
        self.get_data()
        if active:
            self.mouse.position = self.screen_center

    @property
    def active(self):
        """State of this component."""
        return self.update in logic.getCurrentScene().pre_draw

    @active.setter
    def active(self, val):
        pre_draw = logic.getCurrentScene().pre_draw
        if val and self.update not in pre_draw:
            pre_draw.append(self.update)
        elif not val and self.update in pre_draw:
            self.initialized = False
            pre_draw.remove(self.update)

    @property
    def rotation(self):
        """Global body and head orientation."""
        return self.obj.worldOrientation, self.head.worldOrientation

    @rotation.setter
    def rotation(self, val):
        self.obj.worldOrientation = val[0]
        self.head.worldOrientation = val[1]

    def stop(self, reset: bool = False):
        """Stop this component.
        
        :param `reset`: Reset the orientation of objects to their original
        state."""
        self.active = False
        if reset:
            self.reset()
        self.initialized = False

    def disable(self):
        """Set this component to inactive."""
        self.active = False

    def enable(self):
        """Set this component to active."""
        self.active = True

    def reset(self, factor=1):
        """Reset the orientation of the objects in this component to their
        original state.
        
        :param `factor`: Smoothing factor of the reset. If < 1, component will
        be reset smoothly."""
        if factor < 1:
            self.active = False
            if self.reset_factor < 1:
                self.obj.localOrientation = self.obj.localOrientation.lerp(self._defaults[0], factor)
                self.head.localOrientation = self.head.localOrientation.lerp(self._defaults[1], factor)
                self.reset_factor = interpolate(self.reset_factor, 1, factor)
                schedule_callback(self.reset, arg=factor)
            else:
                self.reset_factor = 0
                self.reset()
        else:
            self.obj.localOrientation = self._defaults[0]
            self.head.localOrientation = self._defaults[1]

    def get_data(self):
        """Get data for this component.
        
        Not intended for manual use."""
        self.x = render.getWindowWidth()//2
        self.y = render.getWindowHeight()//2
        self.screen_center = (
            self.x / render.getWindowWidth(),
            self.y / render.getWindowHeight()
        )
        self.center = Vector(self.screen_center)
        self.mouse = logic.mouse

    def update(self):
        """This is executed each frame if component is active."""
        self.get_data()
        if not self.initialized:
            self.mouse.position = self.screen_center
            self.initialized = True
            return
        game_object_x = self.obj
        game_object_y = self.head
        sensitivity = self.sensitivity * 10
        cap_x = self.cap_x
        lowercapX = cap_x[0] * pi / 180
        uppercapX = cap_x[1] * pi / 180
        cap_y = self.cap_y
        lowercapY = cap_y[0] * pi / 180
        uppercapY = cap_y[1] * pi / 180
        invert = self.invert
        smooth = 1 - (self.smoothing * .99)

        mouse_position = Vector(self.mouse.position)
        offset = (mouse_position - self.center) * -0.2

        if invert[1] is False:
            offset.y = -offset.y
        if invert[0] is True:
            offset.x = -offset.x
        offset *= sensitivity

        self._x = offset.x = interpolate(self._x, offset.x, smooth, 0)
        self._y = offset.y = interpolate(self._y, offset.y, smooth, 0)

        if self.use_cap_x:
            objectRotation = game_object_x.localOrientation.to_euler()

            if objectRotation.z + offset.x > uppercapX:
                offset.x = 0
                objectRotation.z = uppercapX
                game_object_x.localOrientation = objectRotation

            if objectRotation.z + offset.x < lowercapX:
                offset.x = 0
                objectRotation.z = lowercapX
                game_object_x.localOrientation = objectRotation

        if not self.axis_lock[0]:
            game_object_x.applyRotation((0, 0, offset.x), self.local)

        rot_axis = 1 - self.front
        if self.use_cap_y:
            objectRotation = game_object_y.localOrientation.to_euler()

            if objectRotation[rot_axis] + offset.y > uppercapY:
                objectRotation[rot_axis] = uppercapY
                game_object_y.localOrientation = objectRotation
                offset.y = 0

            if objectRotation[rot_axis] + offset.y < lowercapY:
                objectRotation[rot_axis] = lowercapY
                game_object_y.localOrientation = objectRotation
                offset.y = 0

        rot = [0, 0, 0]
        rot[1-self.front] = offset.y
        if not self.axis_lock[1]:
            game_object_y.applyRotation((*rot, ), True)
        if (Vector(self.mouse.position) - Vector(self.screen_center)).length > .00001:
        # if self.mouse.position != self.screen_center:
            self.mouse.position = self.screen_center
        self.done = True


class MouseLook(ULMouseLook):
    _deprecated = False



'''TODO: Documentation
'''

from bge import logic
import time
from uplogic.physics import on_collision  # noqa


def on_pre_draw(callback):
    ULEventManager.update_on.append(callback)


def on_post_draw(callback):
    ULEventManager.update_on.append(callback)


def get_event_manager():
    if ULEventManager.update not in ULEventManager.update_on:
        ULEventManager.update_on.append(ULEventManager.update)


def set_update_loop(loop):
    ULEventManager.set_update_on(loop)


class ULEventManager():
    '''Manager for `ULEvent` objects, not inteded for manual use.
    '''
    events = {}
    callbacks = []
    done = []
    update_on = logic.getCurrentScene().post_draw

    @classmethod
    def set_update_on(cls, li):
        if cls.update in cls.update_on:
            cls.update_on.remove(cls.update)
        if cls.update not in li:
            li.append(cls.update)
        cls.update_on = li

    @classmethod
    def update(cls):
        for cb in cls.callbacks.copy():
            cb()

    @classmethod
    def update(cls):
        for cb in cls.callbacks.copy():
            cb()

    @classmethod
    def log(cls):
        if cls.events:
            print('Events:')
            for evt in cls.events:
                print(f'\t{evt}:\t{cls.events[evt].content}')

    @classmethod
    def schedule(cls, cb):
        get_event_manager()
        cls.callbacks.append(cb)

    @classmethod
    def cancel(cls, cb):
        get_event_manager()
        if cb in cls.callbacks:
            cls.callbacks.remove(cb)

    @classmethod
    def bind(cls, cb):
        get_event_manager()
        cls.schedule(cb)

    @classmethod
    def unbind(cls, cb):
        get_event_manager()
        cls.cancel(cb)

    @classmethod
    def register(cls, event):
        get_event_manager()
        cls.events[event.id] = event
        cls.schedule(event.remove)

    @classmethod
    def send(cls, id, content, messenger) -> None:
        get_event_manager()
        ULEvent(id, content, messenger)

    @classmethod
    def receive(cls, id):
        get_event_manager()
        return cls.events.get(id, None)

    @classmethod
    def consume(cls, id):
        get_event_manager()
        return cls.events.pop(id, None)


class ULEvent():
    '''Event generated by `uplogic.events.send()`.

    **Not intended for manual use.**

    :param `id`: Identifier of the event; can be anything, not just `str`.
    :param `content`: This can be used to store data in an event.
    :param `messenger`: Can be used to store an object.
    '''

    def __init__(self, id, content=None, messenger=None):
        self.id = id
        self.content = content
        self.messenger = messenger
        ULEventManager.schedule(self.register)

    def register(self):
        ULEventManager.register(self)
        ULEventManager.cancel(self.register)

    def remove(self):
        ULEventManager.events.pop(self.id, None)
        ULEventManager.cancel(self.remove)


def send(id, content=None, messenger=None) -> None:
    '''Send an event that can be reacted to.

    :param `id`: Identifier of the event; can be anything, not just `str`.
    :param `content`: This can be used to store data in an event.
    :param `messenger`: Can be used to store an object.
    '''
    ULEventManager.send(id, content, messenger)


def receive(id) -> ULEvent:
    '''Check if an event has occured.

    :param `id`: Identifier of the event; can be anything, not just `str`.

    :returns: `ULEvent` with `id`, `content` and `messenger` as attributes.
    '''
    return ULEventManager.receive(id)


def consume(id: str):
    '''Check if an event has occured. This will remove the event.

    :param `id`: Identifier of the event; can be anything, not just `str`.

    :returns: `ULEvent` with `id`, `content` and `messenger` as attributes.
    '''
    return ULEventManager.consume(id, None)


def bind(id, callback) -> None:
    '''Bind a callback to an event.

    :param `id`: Name of the event; can be anything, not just `str`.
    :param `callback`: This callback will be called every time the event is
    triggered.
    '''
    class BoundCallback():
        def __init__(self, id, cb) -> None:
            self.id = id
            self.callback = cb
            ULEventManager.bind(self._check_evt)

        def _check_evt(self):
            evt = receive(self.id)
            if evt:
                self.callback(evt)

        def unbind(self):
            ULEventManager.unbind(self._check_evt)

    return BoundCallback(id, callback)


class ScheduledEvent():
    '''Event generated by `uplogic.events.schedule()`.

    :param `delay`: Delay with which to send the event in seconds.
    :param `id`: Identifier of the event; can be anything, not just `str`
    :param `content`: This can be used to store data in an event.
    :param `messenger`: Can be used to store an object.
    '''

    def __init__(self, delay, id, content, messenger):
        self.time = time.time()
        self.delay = self.time + delay
        self.id = id
        self.content = content
        self.messenger = messenger
        ULEventManager.schedule(self.send_scheduled)

    def send_scheduled(self):
        if time.time() >= self.delay:
            ULEventManager.cancel(self.send_scheduled)
            ULEvent(self.id, self.content, self.messenger)

    def cancel(self):
        ULEventManager.cancel(self.send_scheduled)


def schedule(id: str, delay=0.0, content=None, messenger=None) -> ScheduledEvent:
    '''Send an event that can be reacted to with a delay.

    :param `id`: Name of the event; can be anything, not just `str`. If `id` is callable, `content` can be used as argument.
    :param `content`: This can be used to store data in an event.
    :param `messenger`: Can be used to store an object.
    :param `delay`: Delay with which to send the event in seconds.
    '''
    if callable(id):
        return ScheduledCallback(id, delay, content)
    return ScheduledEvent(delay, id, content, messenger)


class ScheduledCallback():
    '''Event generated by `uplogic.events.schedule_callback()`.

    **Not intended for manual use.**

    :param `cb`: Callback to be evaluated.
    :param `delay`: Delay with which to call the function in seconds.
    :param `arg`: If this is defined, callback will be called with this
    argument.
    '''

    def __init__(self, cb, delay=0.0, arg=None):
        self.time = time.time()
        self.delay = self.time + delay
        self.callback = cb
        self.arg = arg
        ULEventManager.schedule(self.call_scheduled)

    def call_scheduled(self):
        if time.time() >= self.delay:
            ULEventManager.cancel(self.call_scheduled)
            if self.arg is not None:
                self.callback(self.arg)
            else:
                self.callback()

    def cancel(self):
        ULEventManager.cancel(self.call_scheduled)


def schedule_callback(cb, delay=0.0, arg=None) -> ScheduledCallback:
    '''Call a function with a delay. The function can have an argument when
    defined as a keyword.

    Callback cannot return anything.

    :param `cb`: Callback to be evaluated.
    :param `delay`: Delay with which to call the function in seconds.
    :param `arg`: If this is defined, callback will be called with this
    argument.
    '''
    return ScheduledCallback(cb, delay, arg)



from bge import logic
from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid


class ULGamepadTrigger(ULParameterNode):
    def __init__(self, axis=0):
        ULParameterNode.__init__(self)
        self.axis = axis
        self.index = None
        self.sensitivity = None
        self.threshold = None
        self._value = None
        self.VAL = ULOutSocket(self, self.get_value)

    def get_value(self):
        socket = self.get_output('changed')
        if socket is None:
            axis = self.get_input(self.axis)
            if is_invalid(axis):
                return STATUS_WAITING
            index = self.get_input(self.index)
            sensitivity = self.get_input(self.sensitivity)
            threshold = self.get_input(self.threshold)
            if logic.joysticks[index]:
                joystick = logic.joysticks[index]
            else:
                self._x_axis_values = 0
                self._y_axis_values = 0
                return STATUS_WAITING
            if is_invalid(joystick):
                return STATUS_WAITING
            value = (
                joystick.axisValues[4] if axis == 0
                else joystick.axisValues[5]
            )

            if -threshold < value < threshold:
                value = 0
            return self.set_output('trigger', value * sensitivity)
        return socket

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import not_met
from uplogic.utils import debug
from uplogic.utils import interpolate


class ULSlowFollow(ULActionNode):
    def __init__(self, value_type='worldPosition'):
        ULActionNode.__init__(self)
        self.value_type = str(value_type)
        self.condition = NotImplementedError
        self.game_object = None
        self.target = None
        self.factor = None
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        game_object = self.get_input(self.game_object)
        target = self.get_input(self.target)
        attribute = self.get_input(self.value_type)
        factor = self.get_input(self.factor)
        if is_waiting(game_object, target, attribute, factor):
            return
        data = getattr(game_object, attribute)
        t_data = getattr(target, attribute)
        setattr(
            game_object,
            attribute,
            data.lerp(t_data, factor)
        )
        if attribute == 'worldScale':
            game_object.reinstancePhysicsMesh(
                game_object,
                game_object.meshes[0]
            )
        self.done = True



from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid


class ULGetSound(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.sound = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        sound = self.get_input(self.sound)
        if is_invalid(sound):
            return STATUS_WAITING
        return sound

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.ui import Label
from uplogic.utils import not_met
import bpy


class ULCreateUILabel(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.parent = None
        self.rel_pos = None
        self.pos = None
        self.rel_size = None
        self.text = None
        self.text_pos = None
        self.font = None
        self.font_size = None
        self.line_height = None
        self.font_color = None
        self.use_shadow = None
        self.shadow_offset = None
        self.shadow_color = None
        self._widget = None
        self.halign_type = 'left'
        self.valign_type = 'bottom'
        self._done = False
        self.OUT = ULOutSocket(self, self._get_done)
        self.WIDGET = ULOutSocket(self, self._get_widget)

    def _get_done(self):
        return self._done

    def _get_widget(self):
        return self._widget

    def evaluate(self):
        self._done = False
        condition = self.get_input(self.condition)
        self._set_ready()
        if not_met(condition):
            return
        ipt = self.get_input
        parent = ipt(self.parent)
        rel_pos = ipt(self.rel_pos)
        pos = ipt(self.pos)
        text = ipt(self.text)
        font = ipt(self.font)
        font_size = ipt(self.font_size)
        line_height = ipt(self.line_height)
        font_color = ipt(self.font_color)
        use_shadow = ipt(self.use_shadow)
        if use_shadow:
            shadow_offset = ipt(self.shadow_offset)
            shadow_color = ipt(self.shadow_color)
        else:
            shadow_offset = [0, 0]
            shadow_color = [0, 0, 0, 0]
        
        font = bpy.data.fonts[font].filepath.replace('\\', '/') if font else 0

        print(rel_pos)

        self._widget = Label(
            pos=pos,
            relative={'pos': rel_pos},
            halign=self.halign_type,
            valign=self.valign_type,
            text=text,
            font=font,
            font_size=font_size,
            line_height=line_height,
            font_color=font_color,
            shadow=use_shadow,
            shadow_offset=shadow_offset,
            shadow_color=shadow_color
        )
        if parent:
            parent.add_widget(self._widget)
        self._done = True
        



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import not_met
from uplogic.utils.raycasting import raycast
from uplogic.utils.raycasting import raycast_face


class ULRaycast(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.origin = None
        self.destination = None
        self.local: bool = None
        self.property_name: str = None
        self.material: str = None
        self.exclude: str = None
        self.xray: bool = None
        self.custom_dist: bool = None
        self.distance: float = None
        self.visualize: bool = None
        self._result = None
        self._picked_object = None
        self._point = None
        self._normal = None
        self._direction = None
        self._material = None
        self._uv = None
        self.advanced = False
        self.RESULT = ULOutSocket(self, self.get_result)
        self.PICKED_OBJECT = ULOutSocket(self, self.get_picked_object)
        self.POINT = ULOutSocket(self, self.get_point)
        self.NORMAL = ULOutSocket(self, self.get_normal)
        self.DIRECTION = ULOutSocket(self, self.get_direction)
        self.MATERIAL = ULOutSocket(self, self.get_material)
        self.UV = ULOutSocket(self, self.get_uv)
        self.network = None

    def setup(self, network):
        self.network = network

    def get_result(self):
        return self._result

    def get_picked_object(self):
        return self._picked_object

    def get_point(self):
        return self._point

    def get_normal(self):
        return self._normal

    def get_direction(self):
        return self._direction

    def get_material(self):
        return self._material

    def get_uv(self):
        return self._uv

    def evaluate(self):
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_value(False)
            self._normal = None
            self._object = None
            return
        origin = self.get_input(self.origin)
        destination = self.get_input(self.destination)
        local: bool = self.get_input(self.local)
        property_name: str = self.get_input(self.property_name)
        material: str = self.get_input(self.material)
        exclude: str = self.get_input(self.exclude)
        xray: bool = self.get_input(self.xray)
        distance: float = self.get_input(self.distance)
        visualize: bool = self.get_input(self.visualize)

        if is_waiting(origin, destination, local, property_name, distance):
            return
        self._set_ready()
        caster = self.network._owner
        res = (None, None, None)
        cast_f = raycast_face if self.advanced else raycast
        res = cast_f(
            caster,
            origin,
            destination,
            distance if self.get_input(self.custom_dist) else 0,
            property_name,
            material,
            exclude,
            xray,
            local,
            65535,
            visualize,
        )
        self._result = res[0] is not None
        self._picked_object = res[0]
        self._point = res[1]
        self._normal = res[2]
        self._direction = res[3]
        if self.advanced:
            self._material = res[4].material_name[2:] if res[4] else None
            self._uv = res[5]



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import not_met
import bpy


attrs = {
    "show": 'bool_value',
    "color": 'color_value',
    "opacity": 'alpha_value',
    "pos": 'vec2_value',
    "size": 'vec2_value',
    "width": 'float_value',
    "height": 'float_value',
    "use_clipping": 'bool_value',
    "halign": 'str_value',
    "valign": 'str_value',
    "border_width": 'int_value',
    "border_color": 'color_value',
    "orientation": 'str_value',
    "spacing": 'int_value',
    "hover_color": 'color_value',
    "text": 'str_value',
    "font": 'font_value',
    "font_color": 'color_value',
    "font_size": 'int_value',
    "font_opacity": 'alpha_value',
    "line_height": 'float_value',
    "text_halign": 'str_value',
    "text_valign": 'str_value',
    "wrap": 'bool_value',
    "shadow": 'bool_value',
    "shadow_offset": 'vec2_value',
    "shadow_color": 'color_value',
    "texture": 'img_value',
    "icon": 'int_value',
    "rows": 'int_value',
    "cols": 'int_value',
}


class ULSetUIWidgetAttr(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.widget = None
        self.bool_value = None
        self.color_value = None
        self.alpha_value = None
        self.vec2_value = None
        self.str_value = None
        self.int_value = None
        self.float_value = None
        self.font_value = None
        self.img_value = None
        self.widget_attr = 'show'
        self.show = None
        self._done = False
        self.OUT = ULOutSocket(self, self._get_done)

    def _get_done(self):
        return self._done

    def evaluate(self):
        self._done = False
        condition = self.get_input(self.condition)
        self._set_ready()
        if not_met(condition):
            return
        widget = self.get_input(self.widget)
        value = self.get_input(getattr(self, attrs.get(self.widget_attr, 0)))
        if self.widget_attr == 'font':
            value = bpy.data.fonts[value].filepath.replace('\\', '/') if value else 0
        if self.widget_attr == 'texture':
            value = bpy.data.images[value].filepath.replace('\\', '/') if value else 0
        setattr(widget, self.widget_attr, value)
        self._done = True



import json
import toml
import configparser
import os
import bpy


class FileWrapper(dict):
    """Wrapper to allow easy access to files. The wrapper can be saved as a variable to modify and save back the file.\n
    Supports `.ini`, `.json` and `.toml` file formats.

    :param `filepath`: Full path to the file.
    """

    def __init__(self, filepath):
        self.formats = {
            'json': {'read': self._read_json, 'write': self._write_json},
            'ini': {'read': self._read_ini, 'write': self._write_ini},
            'toml': {'read': self._read_toml, 'write': self._write_toml}
        }
        self.filepath = bpy.path.abspath(filepath)
        self.read()

    @property
    def data(self):
        return dict(self)

    @data.setter
    def data(self, data: dict):
        self.clear()
        for key, value in data.items():
            self[key] = value

    def read(self):
        ending = self.filepath.split('.')[-1]
        self.formats[ending]['read']()

    def write(self):
        ending = self.filepath.split('.')[-1]
        self.formats[ending]['write']()

    def _read_json(self):
        with open(self.filepath, 'r') as f:
            self.data = json.load(f)

    def _read_toml(self):
        with open(self.filepath, 'r') as f:
            self.data = toml.load(f)

    def _read_ini(self):
        config = configparser.ConfigParser()
        config.read(self.filepath)
        for section in config.sections():
            self[section] = {}
            for key, val in config[section].items():
                self[section][key] = val

    def _write_json(self):
        path = self.filepath
        if os.path.isfile(path):
            with open(path, 'w') as f:
                json.dump(self, f, indent=2)

    def _write_toml(self):
        path = self.filepath
        if os.path.isfile(path):
            with open(path, 'w') as f:
                toml.dump(self.data, f)

    def _write_ini(self):
        config = configparser.ConfigParser()
        config.read_dict(self)
        with open(self.filepath, 'w') as f:
            config.write(f)


def load_file(filepath) -> FileWrapper:
    """Read a file and keep it stored for later use.\n
    Supports `.ini`, `.json` and `.toml` file formats.

    :param `filepath`: Full path to the file.
    """
    if filepath:
        return FileWrapper(filepath)
    else:
        raise FileNotFoundError(f'File {filepath} could not be opened!')


def write_file(filepath, data) -> None:
    """Write data in form of `dict` back into a file.\n
    Supports `.ini`, `.json` and `.toml` file formats.

    :param `filepath`: Full path to the file.
    :param `data`: `dict` to save back into the file.
    """
    file = FileWrapper(filepath)
    file.data = data
    file.write()


def read_file(filepath) -> dict:
    """Read data in form of `dict` from a file.\n
    Supports `.ini`, `.json` and `.toml` file formats.

    :param `filepath`: Full path to the file.
    """
    return FileWrapper(filepath).data



from bge import logic
from uplogic.nodes import ULConditionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting


class ULMousePressedOn(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.game_object = None
        self.mouse_button = None
        self.OUT = ULOutSocket(self, self.get_changed)

    def get_changed(self):
        socket = self.get_output('changed')
        if socket is None:
            mouse_button = self.get_input(self.mouse_button)
            game_object = self.get_input(self.game_object)
            if is_waiting(mouse_button, game_object):
                return STATUS_WAITING
            if mouse_button is None:
                return STATUS_WAITING
            if is_invalid(game_object):
                return STATUS_WAITING
            mstat = logic.mouse.inputs[mouse_button]
            if not mstat.activated:
                return (False)
            mpos = logic.mouse.position
            camera = logic.getCurrentScene().active_camera
            vec = 10 * camera.getScreenVect(*mpos)
            ray_target = camera.worldPosition - vec
            distance = camera.getDistanceTo(game_object) * 2.0
            t, p, n = self.network.ray_cast(
                camera,
                None,
                ray_target,
                None,
                False,
                distance
            )
            return self.set_output(
                'changed',
                (t == game_object)
            )
        return socket

    def evaluate(self):
        self._set_ready()



from .shader import Filter2D
from mathutils import Vector


glsl = """
uniform sampler2D bgl_RenderedTexture;
in vec4 bgl_TexCoord;
vec2 texcoord = bgl_TexCoord.xy;
uniform vec3 color;

out vec4 fragColor;

vec4 gradient(vec4 coo)
{
    vec4 stripes = coo;
    stripes.r *= color.x;
    stripes.g *= color.y;
    stripes.b *= color.z;
    stripes.a = 1.0;
    return stripes;
}

void main (void) 
{
    vec4 value = texture(bgl_RenderedTexture, texcoord);


//     fragColor = gradient(vec4(clamp(gl_TexCoord[3].s,0.0,1.0)));
    fragColor.rgb = gradient(value).rgb;
    fragColor.a = 1.0;
}
"""

class Levels(Filter2D):

    def __init__(self, color=(1., 1., 1.), idx: int = None) -> None:
        self.settings = {'color': Vector(color)}
        super().__init__(glsl, idx, {'color': self.settings})

    @property
    def color(self):
        return self.settings.get('color', Vector((0, 0, 0)))

    @color.setter
    def color(self, val):
        self.settings['color'] = val



from .widget import Widget
import gpu
import bge, bpy
from math import ceil
from gpu_extras.batch import batch_for_shader


class Image(Widget):

    def __init__(self, pos=[0, 0], size=(100, 100), relative={}, texture=None, halign='left', valign='bottom'):
        super().__init__(pos, size, relative=relative, halign=halign, valign=valign)
        self.texture = texture

    @property
    def texture(self):
        return self._texture
        
    @texture.setter
    def texture(self, val):
        texture = bpy.data.images.get(val)
        if texture:
            self._texture = gpu.texture.from_image(texture)

    def build_shader(self):
        pos = self._draw_pos
        size = self._draw_size
        vertices = self._vertices = (
            (pos[0], pos[1]),
            (pos[0] + size[0], pos[1]),
            (pos[0] + size[0], pos[1] + size[1]),
            (pos[0], pos[1] + size[1])
        )
        self._shader = gpu.shader.from_builtin('2D_IMAGE')
        self._batch = batch_for_shader(
            self._shader, 'TRI_FAN',
            {
                "pos": vertices,
                "texCoord": ((0, 0), (1, 0), (1, 1), (0, 1)),
            },
        )
    
    def draw(self):
        super()._setup_draw()
        self._shader.bind()
        self._shader.uniform_sampler("image", self.texture)
        self._batch.draw(self._shader)
        super().draw()


class Icon(Image):

    def __init__(self, pos=[0, 0], size=(100, 100), relative={}, texture=None, icon=0, rows=1, cols=1, halign='left', valign='bottom'):
        super().__init__(pos, size, relative, texture, halign=halign, valign=valign)
        self.icon = icon
        self.rows = rows
        self.cols = cols

    @property
    def rows(self):
        return self._rows

    @rows.setter
    def rows(self, val):
        if val < 1:
            return
        self._rows = val
        self._row_height = 1 / val

    @property
    def cols(self):
        return self._cols

    @cols.setter
    def cols(self, val):
        if val < 1:
            return
        self._cols = val
        self._col_width = 1 / val

    def build_shader(self):
        pos = self._draw_pos
        size = self._draw_size
        vertices = self._vertices = (
            (pos[0], pos[1]),
            (pos[0] + size[0], pos[1]),
            (pos[0] + size[0], pos[1] + size[1]),
            (pos[0], pos[1] + size[1])
        )
        self._shader = gpu.shader.from_builtin('2D_IMAGE')
        icon = self.icon
        col = icon % self.cols
        col_end = col + 1
        row = ceil((icon + 1) / self.cols) - 1
        row_end = row + 1
        texcoord = (
            (col * self._col_width, 1 - row_end * self._row_height),
            (col_end * self._col_width, 1 - row_end * self._row_height),
            (col_end * self._col_width, 1 - row * self._row_height),
            (col * self._col_width, 1 - row * self._row_height)
        )
        self._batch = batch_for_shader(
            self._shader, 'TRI_FAN',
            {
                "pos": vertices,
                "texCoord": texcoord
            },
        )


from bge.render import drawLine
from uplogic.nodes import ULActionNode
from uplogic.utils import is_invalid
from uplogic.utils import not_met


class ULDrawLine(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.color = None
        self.from_point = None
        self.to_point = None

    def evaluate(self):
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        from_point = self.get_input(self.from_point)
        to_point = self.get_input(self.to_point)
        color = self.get_input(self.color)
        if is_invalid(from_point, to_point, color):
            return
        self._set_ready()
        drawLine(
            from_point,
            to_point,
            [
                color.x,
                color.y,
                color.z,
                1
            ]
        )
        self._set_value(True)



from bge import logic, constraints
from mathutils import Euler
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import debug
from uplogic.utils import is_waiting
from uplogic.utils import not_met
import bpy
import json


class ULLoadGame(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.slot = None
        self.path = ''
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def get_game_vec(self, data):
        return Euler((data['x'], data['y'], data['z']))

    def get_custom_path(self, path):
        if not path.endswith('/') and not path.endswith('json'):
            path = path + '/'
        return path

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        self._set_ready()
        slot = self.get_input(self.slot)
        if is_waiting(slot):
            return
        cust_path = self.get_custom_path(self.path)

        path = (
            logic.expandPath('//Saves/') if self.path == ''
            else cust_path
        )

        scene = logic.getCurrentScene()

        try:
            with open(path + 'save' + str(slot) + '.json') as json_file:
                data = json.load(json_file)
                # for obj in scene.objects:
                #     if obj.name not in data['objects']:
                #         obj.endObject()
                for obj in data['objects']:
                    # print(obj)
                    if obj['name'] in scene.objects:
                        game_obj = scene.objects[obj['name']]
                    else:
                        game_obj = scene.convertBlenderObject(bpy.data.objects[obj['name']])
                        # game_obj = scene.addObject(game_obj)
                        # debug(
                        #     'Could not load Object {}: Not in active Scene!'
                        #     .format(obj['name'])
                        # )
                        # continue

                    lPos = self.get_game_vec(obj['data']['localPosition'])
                    lOri = self.get_game_vec(obj['data']['localOrientation'])
                    lSca = self.get_game_vec(obj['data']['localScale'])

                    wPos = self.get_game_vec(obj['data']['worldPosition'])
                    wOri = self.get_game_vec(obj['data']['worldOrientation'])
                    wSca = self.get_game_vec(obj['data']['worldScale'])

                    game_obj.localPosition = lPos
                    game_obj.localOrientation = lOri.to_matrix()
                    game_obj.localScale = lSca
                    
                    game_obj.worldPosition = wPos
                    game_obj.worldOrientation = wOri.to_matrix()
                    game_obj.worldScale = wSca

                    if obj['type'] == 'rigid_body':
                        linVel = self.get_game_vec(
                            obj['data']['worldLinearVelocity']
                        )
                        angVel = self.get_game_vec(
                            obj['data']['worldAngularVelocity']
                        )
                        game_obj.worldLinearVelocity = linVel
                        game_obj.worldAngularVelocity = angVel

                    if obj['type'] == 'light':
                        energy = obj['data']['energy']
                        game_obj.energy = energy

                    if obj['type'] == 'character':
                        wDir = self.get_game_vec(obj['data']['walkDirection'])
                        (
                            constraints
                            .getCharacter(game_obj)
                            .walkDirection
                        ) = wDir

                    for prop in obj['data']['props']:
                        game_obj[prop['name']] = prop['value']
        except Exception as e:
            debug(
                f'Load Game Node: Could not load saved game on slot {slot}!\n{e}'
            )

        self.done = True



from bge import constraints
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULAddPhysicsConstraint(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.target = None
        self.child = None
        self.name = None
        self.constraint = None
        self.use_world = None
        self.pivot = None
        self.use_limit = None
        self.axis_limits = None
        self.linked_col = None
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        target = self.get_input(self.target)
        child = self.get_input(self.child)
        name = self.get_input(self.name)
        constraint = self.get_input(self.constraint)
        pivot = self.get_input(self.pivot)
        use_limit = self.get_input(self.use_limit)
        use_world = self.get_input(self.use_world)
        axis_limits = self.get_input(self.axis_limits)
        linked_col = self.get_input(self.linked_col)
        if is_invalid(
            target,
            child,
        ):
            return
        if is_waiting(
            name,
            constraint,
            pivot,
            use_limit,
            use_world,
            axis_limits,
            linked_col
        ):
            return
        self._set_ready()
        flag = 0 if linked_col else 128
        if use_world:
            pivot.x -= target.localPosition.x
            pivot.y -= target.localPosition.y
            pivot.z -= target.localPosition.z
        if use_limit:
            target[name] = constraints.createConstraint(
                target.getPhysicsId(),
                child.getPhysicsId(),
                constraint,
                pivot_x=pivot.x,
                pivot_y=pivot.y,
                pivot_z=pivot.z,
                axis_x=axis_limits.x,
                axis_y=axis_limits.y,
                axis_z=axis_limits.z,
                flag=flag
            )
        else:
            target[name] = constraints.createConstraint(
                target.getPhysicsId(),
                child.getPhysicsId(),
                constraint,
                pivot_x=pivot.x,
                pivot_y=pivot.y,
                pivot_z=pivot.z,
                flag=flag
            )
        self.done = True



from .widget import Widget
from bge import render
import gpu
import bge


class Canvas(Widget):

    def __init__(self):
        super().__init__((0, 0), (0, 0))
        self._hover_consumed = False
        self._click_consumed = False
        self.use_clipping = False
        self._to_evaluate = []
        bge.logic.getCurrentScene().post_draw.append(self.draw)

    @property
    def _draw_pos(self):
        return [0, 0]

    @property
    def _draw_size(self):
        return [
            render.getWindowWidth(),
            render.getWindowHeight()
        ]

    @property
    def clipping(self):
        return [
            0,
            render.getWindowWidth(),
            render.getWindowHeight(),
            0
        ]

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, val):
        self._parent = val
        self.pos = self.pos

    def draw(self):
        if not self.show:
            return
        gpu.state.blend_set('ALPHA')
        super().draw()
        self._hover_consumed = False
        while self._to_evaluate:
            self._to_evaluate.pop().update()

    def update(self):
        pass



from uplogic.animation.sequence import Sequence
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULPaySequence(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.mat_name = None
        self.node_name = None
        self.play_mode = None
        self.play_continue = None
        self.frames = None
        self.sequence = None
        self.fps = None
        self.ON_START = ULOutSocket(self, self._get_on_start)
        self.RUNNING = ULOutSocket(self, self._get_running)
        self.ON_FINISH = ULOutSocket(self, self._get_on_finish)
        self.FRAME = ULOutSocket(self, self._get_frame)

    def _get_on_start(self):
        return self.on_start

    def _get_running(self):
        return getattr(self.sequence.running, False)

    def _get_on_finish(self):
        return self.on_finish

    def _get_frame(self):
        return self.frame

    def evaluate(self):
        self.on_finish = False
        self.on_start = False
        condition = self.get_input(self.condition)
        play_continue = self.get_input(self.play_continue)
        if self.sequence:
            if self.sequence.on_finish:
                self.on_finish = True
                if self.sequence.mode < 3:
                    self.sequence = None
        play_mode = self.get_input(self.play_mode)
        frames = self.get_input(self.frames)
        if not_met(condition) and play_mode < 2:
            return
        elif not_met(condition) and self.sequence:
            if not play_continue:
                self.sequence.restart()
            self.sequence.pause()
        elif condition and self.sequence:
            self.sequence.unpause()
        mat_name = self.get_input(self.mat_name)
        node_name = self.get_input(self.node_name)
        fps = self.get_input(self.fps)
        if is_waiting(
            mat_name,
            node_name,
            play_mode,
            frames,
            fps
        ):
            return
        if not self.sequence:
            if play_mode > 2:
                play_mode -= 3
            self.sequence = Sequence(
                mat_name,
                node_name,
                frames.x,
                frames.y,
                fps,
                play_mode
            )



from bge import logic
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket


class ULRestartGame(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        self._set_ready()
        condition = self.get_input(self.condition)
        if condition:
            # logic.getCurrentScene().restart()
            logic.restartGame()
        self.done = True



from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket


class ULGetUIWidgetAttr(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.widget = None
        self.widget_attr = 'show'
        self.BOOL = ULOutSocket(self, self._get_attr)
        self.COLOR = ULOutSocket(self, self._get_attr)
        self.ALPHA = ULOutSocket(self, self._get_attr)
        self.VEC2 = ULOutSocket(self, self._get_attr)
        self.STR = ULOutSocket(self, self._get_attr)
        self.INT = ULOutSocket(self, self._get_attr)
        self.FLOAT = ULOutSocket(self, self._get_attr)
        self.FONT = ULOutSocket(self, self._get_attr)
        self.IMG = ULOutSocket(self, self._get_attr)
        
    def _get_attr(self):
        socket = self.get_output('attr')
        if socket is None:
            widget = self.get_input(self.widget)
            return self.set_output('attr', getattr(widget, self.widget_attr, None))
        return socket

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULConditionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting


class ULNot(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.condition = None
        self.OUT = ULOutSocket(self, self.get_out)

    def get_out(self):
        socket = self.get_output('out')
        if socket is None:
            condition = self.get_input(self.condition)
            if is_waiting(condition):
                return STATUS_WAITING
            return self.set_output(
                'out',
                not condition
            )
        return socket

    def evaluate(self):
        self._set_ready()



from bge import logic
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid


class ULActiveCamera(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.OUT = ULOutSocket(self, self.get_camera)

    def get_camera(self):
        scene = logic.getCurrentScene()
        if is_invalid(scene):
            return STATUS_WAITING
        return scene.active_camera

    def evaluate(self):
        self._set_ready()



from bge import render
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.ui import Cursor, remove_custom_cursor
from uplogic.utils import not_met


class ULSetCustomCursor(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.texture = None
        self.size: list = None
        self.done: bool = None
        self._cursor = None
        self.OUT = ULOutSocket(self, self.get_done)
        self.CURSOR = ULOutSocket(self, self.get_cursor)

    def get_done(self):
        return self.done

    def get_cursor(self):
        return self._cursor

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_ready()
            return
        self._set_ready()
        remove_custom_cursor()
        texture = self.get_input(self.texture)
        size = self.get_input(self.size)
        self._cursor = Cursor(size, texture)
        self.done = True



from typing import Callable
from bge import logic
from bge.types import KX_GameObject as GameObject
from uplogic.utils import debug


class ULCollision():
    """Collision Handler.

    Not intended for manual use."""
    target = None
    point = None
    normal = None
    tap = False
    consumed = False
    active = False
    _objects = []
    _old_objs = []
    done_objs = []

    _deprecated = True

    def __init__(
        self,
        game_object: GameObject,
        callback: Callable,
        prop: str = '',
        mat: str = '',
        tap: bool = False,
        post_call: bool = False
    ):
        if self._deprecated:
            debug('ULCollision class will be renamed to "Collision" in future releases!')
        self.callback: Callable = callback
        self.prop: str = prop
        self.mat: str = mat
        self.tap: bool = tap
        self.post_call = post_call
        self.game_object: GameObject = game_object
        self.register()

    def collision(self, obj, point, normal):
        self._objects.append(obj)
        material = self.mat
        prop = self.prop
        bo = obj.blenderObject
        if material:
            if material not in [
                slot.material.name for
                slot in
                bo.material_slots
            ]:
                return
        if prop:
            if prop not in obj.getPropertyNames():
                return
        if self.tap and self.consumed:
            self.active = True
            return

        self.active = True
        if obj not in self.done_objs:
            if self.game_object.collisionGroup & obj.collisionMask and self.game_object.collisionMask & obj.collisionGroup:
                self.callback(obj, point, normal)
            self.done_objs.append(obj)

    def reset(self):
        if self.post_call:
            for obj in self._old_objs:
                if obj not in self.done_objs:
                    self.callback(None, None, None)
        self._old_objs = self.done_objs
        self.done_objs = []
        if not self.consumed and self.active:
            self.consumed = True
        elif self.consumed and not self.active:
            self.consumed = False
        self.active = False

    def register(self):
        if self.collision not in self.game_object.collisionCallbacks:
            self.game_object.collisionCallbacks.append(self.collision)
        logic.getCurrentScene().pre_draw.append(self.reset)

    def remove(self):
        self.game_object.collisionCallbacks.remove(self.collision)
        logic.getCurrentScene().pre_draw.remove(self.reset)


class Collision(ULCollision):
    _deprecated = False


def on_collision(
    obj: GameObject,
    callback: Callable,
    prop: str = '',
    material: str = '',
    tap: bool = False,
    post_call: bool = False
) -> Collision:
    """Bind a callback to an object's collision detection.

    :param `obj`: Object whose collision detection will be monitored.
    :param `callback`: Callback to be called when collision occurs. Must have arguments `(obj, point, normal)`.
    :param `prop`: Only look for objects that have this property.
    :param `material`: Only look for objects that have this material applied.
    :param `tap`: Only validate the first frame of the collision.
    """
    return Collision(obj, callback, prop, material, tap, post_call)



from bge import logic
from uplogic.nodes import ULConditionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting


class ULMouseOver(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.game_object = None
        self._mouse_entered_status = False
        self._mouse_exited_status = False
        self._mouse_over_status = False
        self._point = None
        self._normal = None
        self.MOUSE_ENTERED = ULOutSocket(self, self._get_mouse_entered)
        self.MOUSE_EXITED = ULOutSocket(self, self._get_mouse_exited)
        self.MOUSE_OVER = ULOutSocket(self, self._get_mouse_over)
        self.POINT = ULOutSocket(self, self._get_point)
        self.NORMAL = ULOutSocket(self, self._get_normal)
        self._last_target = None

    def _get_mouse_entered(self):
        return self._mouse_entered_status

    def _get_mouse_exited(self):
        return self._mouse_exited_status

    def _get_mouse_over(self):
        return self._mouse_over_status

    def _get_point(self):
        return self._point

    def _get_normal(self):
        return self._normal

    def evaluate(self):
        game_object = self.get_input(self.game_object)
        if is_waiting(game_object):
            return
        self._set_ready()
        if is_invalid(game_object):
            self._mouse_entered_status = False
            self._mouse_exited_status = False
            self._mouse_over_status = False
            self._point = None
            self._normal = None
            return
        scene = game_object.scene
        camera = scene.active_camera
        distance = 2.0 * camera.getDistanceTo(game_object)
        mouse = logic.mouse
        mouse_position = mouse.position
        vec = 10.0 * camera.getScreenVect(*mouse_position)
        ray_target = camera.worldPosition - vec
        target, point, normal = self.network.ray_cast(
            camera,
            None,
            ray_target,
            None,
            False,
            distance
        )
        if not (target is self._last_target):  # mouse over a new object
            # was target, now it isn't -> exited
            if self._last_target is game_object:
                self._mouse_exited_status = True
                self._mouse_over_status = False
                self._mouse_entered_status = False
                self._point = None
                self._normal = None
            # wasn't target, now it is -> entered
            elif (target is game_object):
                self._mouse_entered_status = True
                self._mouse_over_status = False
                self._mouse_exited_status = False
                self._point = point
                self._normal = normal
            self._last_target = target
        else:  # the target has not changed
            # was target, still target -> over
            if self._last_target is game_object:
                self._mouse_entered_status = False
                self._mouse_exited_status = False
                self._mouse_over_status = True
                self._point = point
                self._normal = normal
            else:  # wans't target, still isn't target -> clear
                self._mouse_entered_status = False
                self._mouse_exited_status = False
                self._mouse_over_status = False
                self._point = None
                self._normal = None



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.ui import Canvas
from uplogic.utils import not_met


class ULCreateUICanvas(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self._canvas = None
        self._done = False
        self.OUT = ULOutSocket(self, self._get_done)
        self.CANVAS = ULOutSocket(self, self._get_canvas)

    def _get_done(self):
        return self._done

    def _get_canvas(self):
        return self._canvas

    def evaluate(self):
        self._done = False
        condition = self.get_input(self.condition)
        self._set_ready()
        if not_met(condition):
            return
        self._canvas = Canvas()
        self._done = True
        



from uplogic.animation import ULActionSystem
from uplogic.data import GlobalDB
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULStopAction(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.action_layer = None
        self.done = None
        self.act_system = self.get_act_system()
        self.OUT = ULOutSocket(self, self.get_done)
        
    def get_act_system(self):
        act_systems = GlobalDB.retrieve('uplogic.animation')
        if act_systems.check('default'):
            return act_systems.get('default')
        else:
            return ULActionSystem('default')

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_ready()
            return
        game_object = self.get_input(self.game_object)
        action_layer = self.get_input(self.action_layer)
        if is_waiting(game_object, action_layer):
            return
        self._set_ready()
        if is_invalid(game_object):
            return
        if is_invalid(action_layer):
            return

        action = self.act_system.get_layer(game_object, action_layer)
        if action is not None:
            self.act_system.remove(action)
        self.done = True



from bge import logic
from bge.constraints import getCharacter
from bge.types import KX_GameObject as GameObject
from mathutils import Vector
from uplogic.utils import debug
import bpy


class ULCharacter():
    _deprecated = True

    def __init__(self, owner: GameObject) -> None:
        if self._deprecated:
            debug('ULCharacter class will be renamed to "Character" in future releases!')
        self.owner = owner
        self.wrapper = getCharacter(owner)
        self._old_position = owner.worldPosition.copy()
        self.velocity = Vector((0, 0, 0))
        self.is_walking = False
        self._on_ground = self.wrapper.onGround
        self.landed = False
        self.start_falling = False
        self.speed = 1
        self._phys_step = bpy.data.scenes[logic.getCurrentScene().name].game_settings.physics_step_sub
        logic.getCurrentScene().pre_draw.append(self.reset)

    def reset(self):
        self._velocity = (self.owner.worldPosition - self._old_position) / 10
        self._old_position = self.owner.worldPosition.copy()
        self.landed = not self._on_ground and self.on_ground
        self.start_falling = self._on_ground and not self.on_ground
        self._on_ground = self.on_ground
        if not self.is_walking:
            self.walk = Vector((0, 0, 0))
        self.is_walking = False

    def destroy(self):
        logic.getCurrentScene().pre_draw.remove(self.reset)

    @property
    def on_ground(self) -> bool:
        return self.wrapper.onGround

    @on_ground.setter
    def on_ground(self, value):
        debug('ULCharacter.on_ground is Read-Only!')

    @property
    def max_jumps(self) -> int:
        return self.wrapper.maxJumps

    @max_jumps.setter
    def max_jumps(self, value):
        self.wrapper.maxJumps = value

    @property
    def gravity(self) -> Vector:
        return self.wrapper.gravity

    @gravity.setter
    def gravity(self, value):
        self.wrapper.gravity = value

    @property
    def slope(self) -> Vector:
        return self.wrapper.maxSlope

    @slope.setter
    def slope(self, value):
        self.wrapper.maxSlope = value

    @property
    def jump_count(self) -> int:
        return self.wrapper.jumpCount

    @jump_count.setter
    def jump_count(self, value):
        debug('Character.jump_count is Read-Only!')

    @property
    def walk(self) -> Vector:
        return ((self.wrapper.walkDirection @ self.owner.worldOrientation) * self._phys_step) / self.speed

    @walk.setter
    def walk(self, value):
        self.is_walking = True
        self.wrapper.walkDirection = ((self.owner.worldOrientation @ value) / self._phys_step) * self.speed

    @property
    def velocity(self) -> Vector:
        return self._velocity

    @velocity.setter
    def velocity(self, value):
        self.wrapper.setVelocity(value, 1, False)

    @property
    def jump_force(self) -> float:
        return self.jump_force

    @jump_force.setter
    def jump_force(self, value):
        self.wrapper.jumpSpeed = value

    @property
    def fall_speed(self) -> float:
        return self.wrapper.fallSpeed

    @fall_speed.setter
    def fall_speed(self, value):
        self.wrapper.fallSpeed = value

    def move(self, direction=Vector((0, 0, 0)), local=True):
        self.is_walking = True
        self.wrapper.walkDirection = self.owner.worldOrientation @ direction * self.speed if local else direction * self.speed

    def jump(self):
        self.wrapper.jump()


class Character(ULCharacter):
    _deprecated = False



from mathutils import Euler
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode


class ULEuler(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.input_x = None
        self.input_y = None
        self.input_z = None
        self.OUTV = ULOutSocket(self, self.get_out_v)

    def get_out_v(self):
        socket = self.get_output('out_v')
        if socket is None:
            e = Euler()
            x = self.get_input(self.input_x)
            y = self.get_input(self.input_y)
            z = self.get_input(self.input_z)
            if x is not None:
                e.x = x
            if y is not None:
                e.y = y
            if z is not None:
                e.z = z
            return self.set_output('out_v', e)
        return socket

    def evaluate(self):
        self._set_ready()



import gpu
from gpu_extras.batch import batch_for_shader
from bge import render


class Widget():
    '''TODO: Documentation
    '''
    def __init__(self, pos=(0, 0), size=(0, 0), bg_color=(0, 0, 0, 0), relative={}, halign='left', valign='bottom'):
        self.halign = halign
        self.valign = valign
        self._parent = None
        self._show = True
        self._pos = [0, 0]
        self._children: list[Widget] = []
        self.relative = relative
        self._rebuild = True 
        self.size = size
        self.pos = pos
        self.bg_color = bg_color
        self._vertices = ((0, 0), (0, 0), (0, 0), (0, 0))
        self._clipped = [0, 0]
        self.use_clipping = False
        self.build_shader()
        self.z = 0
        self.start()

    def toggle(self, *args):
        self.show = not self.show

    @property
    def show(self):
        return self._show

    @show.setter
    def show(self, val):
        if val != self._show:
            self._show = val
            if val:
                self._rebuild = True
                for child in self.children:
                    child.pos = child.pos

    @property
    def _children_reversed(self):
        return self.children.__reversed__()

    @property
    def canvas(self):
        pa = self
        while pa.parent is not None:
            pa = pa.parent
        return pa

    @property
    def _recurse(self):
        widgets = [self]
        for w in self.children:
            widgets.extend(w._recurse)
        return widgets

    @property
    def childrenRecursive(self):
        widgets = []
        for w in self.children:
            widgets.extend(w._recurse)
        return widgets

    @property
    def children(self):
        return self._children

    @children.setter
    def children(self, val):
        self._children = val

    @property
    def bg_color(self):
        return self._bg_color

    @bg_color.setter
    def bg_color(self, val):
        val = list(val)
        self._bg_color = val

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, val):
        if self.parent is not val and self.parent:
            self.parent.remove_widget(self)
        if self.use_clipping is None:
            self.use_clipping = val.use_clipping
        self._parent = val
        self.pos = self.pos
        self.size = self.size
        self.build_shader()

    @property
    def pos_abs(self):
        pos = self._vertices[0]
        return [
            pos[0] - self._clipped[0],
            pos[1] - self._clipped[1]
        ]

    @property
    def pos(self):
        return self._pos

    @pos.setter
    def pos(self, val):
        self._pos = list(val)
        if not self.show:
            return
        if self.parent and self.show:
            self._rebuild = True
        for child in self.children:
            child.pos = child.pos

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, val):
        self._size = list(val)
        if not self.show:
            return
        if self.parent and self.show:
            self._rebuild = True
        for child in self.children:
            child.pos = child.pos

    @property
    def use_clipping(self):
        return self._use_clipping

    @use_clipping.setter
    def use_clipping(self, val):
        self._use_clipping = val
        for widget in self.childrenRecursive:
            widget._use_clipping = val
        if self.parent and self.show:
            self._rebuild = True

    @property
    def width(self):
        return self.size[0]

    @width.setter
    def width(self, val):
        self.size[0] = val
        if self.parent and self.show:
            self._rebuild = True

    @property
    def height(self):
        return self.size[1]

    @height.setter
    def height(self, val):
        self.size[1] = val
        if self.parent and self.show:
            self._rebuild = True

    @property
    def opacity(self):
        return self.bg_color[3]

    @opacity.setter
    def opacity(self, val):
        self.bg_color[3] = val
        if self.parent and self.show:
            self._rebuild = True

    @property
    def child_offset(self):
        return [0, 0]

    @property
    def clipping(self):
        pdpos = self.parent._draw_pos
        pdsize = self.parent._draw_size
        return [
            pdpos[0],
            pdpos[0] + pdsize[0],
            pdpos[1] + pdsize[1],
            pdpos[1]
        ]

    @property
    def _draw_pos(self):
        if self.parent is None:
            return [0, 0]
        inherit_pos = self.parent.pos_abs if self.parent else [0, 0]
        pdsize = self.parent._draw_size
        pos = [
            self.pos[0] * pdsize[0],
            self.pos[1] * pdsize[1]
        ] if self.relative.get('pos') else self.pos
        offset = [0, 0]
        dsize = self._draw_size
        if self.halign == 'center':
            offset[0] += dsize[0] * .5
        elif self.halign == 'right':
            offset[0] += dsize[0]
        if self.valign == 'center':
            offset[1] += dsize[1] * .5
        elif self.valign == 'top':
            offset[1] += dsize[1]
        return [pos[0] + inherit_pos[0] - offset[0], pos[1] + inherit_pos[1] - offset[1]]

    @property
    def _draw_size(self):
        if self.parent is None:
            return self.size
        if self.relative.get('size'):
            return [
                self.size[0] * self.parent._draw_size[0],
                self.size[1] * self.parent._draw_size[1]
            ]
        return self.size

    def start(self):
        pass

    def build_shader(self):
        if self.parent is None:
            return
        pos = self._draw_pos
        size = self._draw_size
        x0 = [pos[0], pos[1]]
        x1 = [pos[0] + size[0], pos[1]]
        y0 = [pos[0], pos[1] + size[1]]
        y1 = [pos[0] + size[0], pos[1] + size[1]]
        v = [x0, x1, y0, y1]
        if self.parent.use_clipping:
            clip = self.clipping
            for vert in v:
                if vert[0] < clip[0]:
                    self._clipped[0] = clip[0] - vert[0]
                    vert[0] = clip[0]
                elif vert[0] > clip[1]:
                    vert[0] = clip[1]
                if vert[1] < clip[3]:
                    self._clipped[1] = clip[3] - vert[1]
                    vert[1] = clip[3]
                elif vert[1] > clip[2]:
                    vert[1] = clip[2]
        vertices = self._vertices = (
            x0,
            x1,
            y1,
            y0
        )
        indices = (
            (0, 1, 2), (2, 3, 0)
        )
        self._shader = gpu.shader.from_builtin('2D_UNIFORM_COLOR')
        self._batch = batch_for_shader(self._shader, 'TRIS', {"pos": vertices}, indices=indices)
        self._batch_line = batch_for_shader(self._shader, 'LINE_LOOP', {"pos": vertices})
        self._batch_points = batch_for_shader(self._shader, 'POINTS', {"pos": vertices})

    def _setup_draw(self):
        if self._rebuild is True:
            self.build_shader()
            self._rebuild = False

    def draw(self):
        """This is called each frame.
        """
        gpu.state.blend_set('ALPHA')
        self.canvas._to_evaluate.append(self)
        for widget in self.children:
            if widget.show:
                widget.draw()

    def update(self):
        pass

    def add_widget(self, widget):
        '''Add a `Widget`.

        :param `action`: `Widget` to add.
        '''
        if widget not in self.children:
            self.children.append(widget)
            self.canvas.set_z(-1)
            widget.parent = self
        self.children = sorted(self.children, key=lambda widget: widget.z, reverse=False)

    def set_z(self, z):
        z += 1
        self.z = z
        for c in self.children:
            z = c.set_z(z)
        return z

    def remove_widget(self, widget):
        '''Remove a `Widget`.

        :param `action`: `Widget` which to remove.
        '''
        if widget in self.children:
            self.children.remove(widget)
            widget.parent = None

    def clear(self):
        to_remove = self.children.copy()
        for child in to_remove:
            self.remove_widget(child)
